# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Specific handling for some HTML elements, especially replaced elements.

Replaced elements (eg. <img> elements) are rendered externally and behave
as an atomic opaque box in CSS. They may or may not have intrinsic dimensions.

"""

from __future__ import division, unicode_literals

from .formatting_structure import boxes
from .utils import get_url_attribute
from .compat import xrange


# Maps HTML tag names to function taking an HTML element and returning a Box.
HTML_HANDLERS = {}


def handle_element(document, element, box):
    """Handle HTML elements that need special care.

    :returns: a (possibly empty) list of boxes.
    """
    if box.element_tag in HTML_HANDLERS:
        return HTML_HANDLERS[element.tag](document, element, box)
    else:
        return [box]


def handler(tag):
    """Return a decorator registering a function handling ``tag`` elements."""
    def decorator(function):
        """Decorator registering a function handling ``tag`` elements."""
        HTML_HANDLERS[tag] = function
        return function
    return decorator


def is_block_level(box):
    """Tell wether ``box`` is supposed to be block level.

    Return ``True`` if the element is block-level, ``False`` if it is
    inline-level, and raise ValueError if it is neither.

    """
    display = box.style.display

    if display in ('block', 'list-item', 'table'):
        return True
    elif display in ('inline', 'inline-table', 'inline-block'):
        return False
    else:
        raise ValueError('Unsupported display: ' + display)


def make_replaced_box(element, box, image):
    """Wrap an image in a replaced box.

    That box is either block-level or inline-level, depending on what the
    element should be.

    """
    if is_block_level(box):
        type_ = boxes.BlockReplacedBox
    else:
        type_ = boxes.InlineReplacedBox
    return type_(element.tag, element.sourceline, box.style, image)


def make_text_box(element, box, text):
    """Make a text box.

    If the element should be block-level, wrap it in a block box.

    """
    text_box = boxes.TextBox(element.tag, element.sourceline,
                             box.style.inherit_from(), text)
    if is_block_level(box):
        type_ = boxes.BlockBox
    else:
        type_ = boxes.InlineBox
    return type_(element.tag, element.sourceline,
                 box.style, [text_box])


@handler('img')
def handle_img(document, element, box):
    """Handle ``<img>`` elements, return either an image or the alt-text.

    See: http://www.w3.org/TR/html5/embedded-content-1.html#the-img-element

    """
    src = get_url_attribute(element, 'src')
    alt = element.get('alt')
    if src:
        image = document.get_image_from_uri(src)
        if image is not None:
            return [make_replaced_box(element, box, image)]
        else:
            # Invalid image, use the alt-text.
            if alt:
                return [make_text_box(element, box, alt)]
            elif alt == '':
                # The element represents nothing
                return []
            else:
                assert alt is None
                # TODO: find some indicator that an image is missing.
                # For now, just remove the image.
                return []
    else:
        if alt:
            return [make_text_box(element, box, alt)]
        else:
            return []


@handler('embed')
def handle_embed(document, element, box):
    """Handle ``<embed>`` elements, return either an image or nothing.

    See: http://www.w3.org/TR/html5/the-iframe-element.html#the-embed-element

    """
    src = get_url_attribute(element, 'src')
    type_ = element.get('type', '').strip()
    if src:
        image = document.get_image_from_uri(src, type_)
        if image is not None:
            return [make_replaced_box(element, box, image)]
    # No fallback.
    return []


@handler('object')
def handle_object(document, element, box):
    """Handle ``<object>`` elements, return either an image or the fallback
    content.

    See: http://www.w3.org/TR/html5/the-iframe-element.html#the-object-element

    """
    data = get_url_attribute(element, 'data')
    type_ = element.get('type', '').strip()
    if data:
        image = document.get_image_from_uri(data, type_)
        if image is not None:
            return [make_replaced_box(element, box, image)]
    # The element’s children are the fallback.
    return [box]


def integer_attribute(element, box, name, minimum=1):
    """Read an integer attribute from the HTML element and set it on the box.

    """
    value = element.get(name, '').strip()
    if value:
        try:
            value = int(value)
        except ValueError:
            pass
        else:
            if value >= minimum:
                setattr(box, name, value)


@handler('colgroup')
def handle_colgroup(_document, element, box):
    """Handle the ``span`` attribute."""
    if isinstance(box, boxes.TableColumnGroupBox):
        if any(child.tag == 'col' for child in element):
            box.span = None  # sum of the children’s spans
        else:
            integer_attribute(element, box, 'span')
            box.children = (
                boxes.TableColumnBox.anonymous_from(box, [])
                for _i in xrange(box.span))
    return [box]


@handler('col')
def handle_col(_document, element, box):
    """Handle the ``span`` attribute."""
    if isinstance(box, boxes.TableColumnBox):
        integer_attribute(element, box, 'span')
        if box.span > 1:
            # Generate multiple boxes
            # http://lists.w3.org/Archives/Public/www-style/2011Nov/0293.html
            return [box.copy() for _i in xrange(box.span)]
    return [box]


@handler('th')
@handler('td')
def handle_td(_document, element, box):
    """Handle the ``colspan``, ``rowspan`` attributes."""
    if isinstance(box, boxes.TableCellBox):
        # HTML 4.01 gives special meaning to colspan=0
        # http://www.w3.org/TR/html401/struct/tables.html#adef-rowspan
        # but HTML 5 removed it
        # http://www.w3.org/TR/html5/tabular-data.html#attr-tdth-colspan
        # rowspan=0 is still there though.
        integer_attribute(element, box, 'colspan')
        integer_attribute(element, box, 'rowspan', minimum=0)
    return [box]



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Various utils.

"""

from __future__ import division, unicode_literals

import io
import base64

from cssutils.helper import path2url

from . import VERSION
from .logger import LOGGER
from .compat import (
    urljoin, urlparse, unquote_to_bytes, urlopen_contenttype, Request,
    parse_email)


HTTP_USER_AGENT = 'WeasyPrint/%s http://weasyprint.org/' % VERSION


def get_url_attribute(element, key):
    """Get the URL corresponding to the ``key`` attribute of ``element``.

    The retrieved URL is absolute, even if the URL in the element is relative.

    """
    attr_value = element.get(key)
    if attr_value:
        attr_value = attr_value.strip()
        if attr_value:
            return urljoin(element.base_url, attr_value)


def ensure_url(string):
    """Get a ``scheme://path`` URL from ``string``.

    If ``string`` looks like an URL, return it unchanged. Otherwise assume a
    filename and convert it to a ``file://`` URL.

    """
    if urlparse(string).scheme:
        return string
    else:
        return path2url(string.encode('utf8'))


def parse_data_url(url):
    """Decode URLs with the 'data' stream. urllib can handle them
    in Python 2, but that is broken in Python 3.

    Inspired from the Python 2.7.2’s urllib.py.

    """
    # syntax of data URLs:
    # dataurl   := "data:" [ mediatype ] [ ";base64" ] "," data
    # mediatype := [ type "/" subtype ] *( ";" parameter )
    # data      := *urlchar
    # parameter := attribute "=" value
    try:
        header, data = url.split(',', 1)
    except ValueError:
        raise IOError('bad data URL')
    header = header[5:]  # len('data:') == 5
    if header:
        semi = header.rfind(';')
        if semi >= 0 and '=' not in header[semi:]:
            content_type = header[:semi]
            encoding = header[semi+1:]
        else:
            content_type = header
            encoding = ''
        message = parse_email('Content-type: ' + content_type)
        mime_type = message.get_content_type()
        charset = message.get_content_charset()
    else:
        mime_type = 'text/plain'
        charset = 'US-ASCII'
        encoding = ''

    if encoding == 'base64':
        data = data.encode('ascii')
        data = base64.decodestring(data)
    else:
        data = unquote_to_bytes(data)

    return io.BytesIO(data), mime_type, charset


def urlopen(url):
    """Fetch an URL and return ``(file_like, mime_type, charset)``.

    It is the caller’s responsability to call ``file_like.close()``.
    """
    if url.startswith('data:'):
        return parse_data_url(url)
    else:
        return urlopen_contenttype(Request(url,
            headers={'User-Agent': HTTP_USER_AGENT}))


def urllib_fetcher(url):
    """URL fetcher for cssutils.

    This fetcher is based on urllib instead of urllib2, since urllib has
    support for the "data" URL scheme.

    """
    file_like, mime_type, charset = urlopen(url)
    if mime_type != 'text/css':
        LOGGER.warn('Expected `text/css` for stylsheet at %s, got `%s`',
                    url, mime_type)
        return None
    content = file_like.read()
    file_like.close()
    return charset, content


class cached_property(object):
    """A decorator that converts a function into a lazy property. The
    function wrapped is called the first time to retrieve the result
    and then that calculated result is used the next time you access
    the value.

    Stolen from Werkzeug:
    https://github.com/mitsuhiko/werkzeug/blob/7b8d887d33/werkzeug/utils.py#L28

    """

    def __init__(self, func):
        self.__name__ = func.__name__
        self.__module__ = func.__module__
        self.__doc__ = func.__doc__
        self.func = func

    def __get__(self, obj, type=None):
        if obj is None:
            return self
        missing = object()
        value = obj.__dict__.get(self.__name__, missing)
        if value is missing:
            value = self.func(obj)
            obj.__dict__[self.__name__] = value
        return value



from __future__ import division, unicode_literals

import functools
from io import BytesIO
import cairo



BOX_ATTRIBUTES = '''
    position_x position_y
    width height
    margin_top margin_bottom margin_left margin_right
    padding_top padding_bottom padding_left padding_right
    border_top_width border_bottom_width border_left_width border_right_width
'''.split()

PAGE_BOX_ATTRIBUTES = BOX_ATTRIBUTES + ['outer_width', 'outer_height']


def serialize(document):
    """
    Serialize a WeasyPrint Document object to JSON-compatible data types
    for in-browser rendering.
    """
    return {'pages': [
        serialize_page_box(document, page)
        for page in document.pages]}


def serialize_box(document, box, attributes=BOX_ATTRIBUTES):
    rv = {name: getattr(box, name) for name in attributes}
    rv['style'] = serialize_style(document, box.style)

    children = getattr(box, 'children', None)
    if children is not None:
        rv['children'] = [serialize_box(document, child) for child in children]

    marker = getattr(box, 'outside_list_marker', None)
    if marker is not None:
        rv['children'].append(serialize_box(document, marker))

    text = getattr(box, 'text', None)
    if text is not None:
        rv['text'] = text

    replacement = getattr(box, 'replacement', None)
    if replacement is not None:
        width = int(box.width)
        height = int(box.height)
        rv['width'] = width
        rv['height'] = height
        rv['image'] = serialize_replacement(replacement, width, height)

    return rv


def serialize_replacement(replacement, scaled_width=None, scaled_height=None):
    pattern, width, height = replacement

    if scaled_width is None:
        scaled_width = width
    if scaled_height is None:
        scaled_height = height

    surface = cairo.ImageSurface(
        cairo.FORMAT_ARGB32, scaled_width, scaled_height)
    context = cairo.Context(surface)
    context.scale(scaled_width / width, scaled_height / height)
    context.set_source(pattern)
    context.paint()
    file_like = BytesIO()
    surface.write_to_png(file_like)
    return file_like.getvalue().encode('base64').replace('\n', '')


serialize_page_box = functools.partial(
    serialize_box, attributes=PAGE_BOX_ATTRIBUTES)

def serialize_style(document, style):
    style = style.as_dict()
    for key, value in style.iteritems():
        if getattr(value, 'type', None) == 'COLOR_VALUE':
            style[key] = value.cssText
    bg_image = style['background_image']
    if bg_image == 'None':
        bg_image = None
    else:
        bg_image = document.get_image_from_uri(style['background_image'])
        if bg_image is not None:
            bg_image = serialize_replacement(bg_image)
    style['background_image'] = bg_image
    style['background_position'] = [
        getattr(v, 'cssText', v)
        for v in style['background_position']]
    style['text_decoration'] = ' '.join(style['text_decoration']) or 'none'
    return style



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Functions laying out tables.

"""

from __future__ import division, unicode_literals

from ..compat import xrange
from ..logger import LOGGER
from ..formatting_structure import boxes
from .percentages import resolve_percentages, resolve_one_percentage


def table_layout(document, table, max_position_y, containing_block,
                 device_size, page_is_empty):
    """Layout for a table box.

    For now only the fixed layout and separate border model are supported.

    """
    # Avoid a circular import
    from .blocks import block_level_height

    column_widths = table.column_widths

    border_spacing_x, border_spacing_y = table.style.border_spacing
    # TODO: reverse this for direction: rtl
    column_positions = []
    position_x = table.content_box_x()
    rows_x = position_x + border_spacing_x
    for width in column_widths:
        position_x += border_spacing_x
        column_positions.append(position_x)
        position_x += width
    rows_width = position_x - rows_x

    # Layout for row groups, rows and cells
    position_y = table.content_box_y() + border_spacing_y
    initial_position_y = position_y
    new_table_children = []
    for group in table.children:
        resolve_percentages(group, containing_block=table)
        group.position_x = rows_x
        group.position_y = position_y
        group.width = rows_width
        new_group_children = []
        # For each rows, cells for which this is the last row (with rowspan)
        ending_cells_by_row = [[] for row in group.children]
        for row in group.children:
            resolve_percentages(row, containing_block=table)
            row.position_x = rows_x
            row.position_y = position_y
            row.width = rows_width
            # Place cells at the top of the row and layout their content
            new_row_children = []
            for cell in row.children:
                spanned_widths = column_widths[cell.grid_x:][:cell.colspan]
                # In the fixed layout the grid width is set by cells in
                # the first row and column elements.
                # This may be less than the previous value  of cell.colspan
                # if that would bring the cell beyond the grid width.
                cell.colspan = len(spanned_widths)
                if cell.colspan == 0:
                    # The cell is entierly beyond the grid width, remove it
                    # entierly. Subsequent cells in the same row have greater
                    # grid_x, so they are beyond too.
                    cell_index = row.children.index(cell)
                    ignored_cells = row.children[cell_index:]
                    LOGGER.warn('This table row has more columns than '
                                'the table, ignored %i cells: %r',
                                len(ignored_cells), ignored_cells)
                    break
                resolve_percentages(cell, containing_block=table)
                cell.position_x = column_positions[cell.grid_x]
                cell.position_y = row.position_y
                cell.margin_top = 0
                cell.margin_left = 0
                cell.width = 0
                borders_plus_padding = cell.border_width()  # with width==0
                cell.width = (
                    sum(spanned_widths)
                    + border_spacing_x * (cell.colspan - 1)
                    - borders_plus_padding)
                # The computed height is a minimum
                computed_cell_height = cell.height
                cell.height = 'auto'
                cell, _, _, _, _ = block_level_height(
                    document, cell,
                    max_position_y=float('inf'),
                    skip_stack=None,
                    device_size=device_size,
                    page_is_empty=True)
                if computed_cell_height != 'auto':
                    cell.height = max(cell.height, computed_cell_height)
                new_row_children.append(cell)

            row = row.copy_with_children(new_row_children)
            new_group_children.append(row)

            # Table height algorithm
            # http://www.w3.org/TR/CSS21/tables.html#height-layout

            # cells with vertical-align: baseline
            baseline_cells = []
            for cell in row.children:
                vertical_align = cell.style.vertical_align
                if vertical_align in ('top', 'middle', 'bottom'):
                    cell.vertical_align = vertical_align
                else:
                    # Assume 'baseline' for any other value
                    cell.vertical_align = 'baseline'
                    cell.baseline = cell_baseline(cell)
                    baseline_cells.append(cell)
            if baseline_cells:
                row.baseline = max(cell.baseline for cell in baseline_cells)
                for cell in baseline_cells:
                    if cell.baseline != row.baseline:
                        add_top_padding(cell, row.baseline - cell.baseline)
            else:
                row.baseline = None

            # row height
            for cell in row.children:
                ending_cells_by_row[cell.rowspan - 1].append(cell)
            ending_cells = ending_cells_by_row.pop(0)
            if ending_cells:  # in this row
                row_bottom_y = max(
                    cell.position_y + cell.border_height()
                    for cell in ending_cells)
                row.height = row_bottom_y - row.position_y
            else:
                row_bottom_y = row.position_y
                row.height = 0

            # Add extra padding to make the cells the same height as the row
            for cell in ending_cells:
                cell_bottom_y = cell.position_y + cell.border_height()
                extra = row_bottom_y - cell_bottom_y
                if cell.vertical_align == 'bottom':
                    add_top_padding(cell, extra)
                elif cell.vertical_align == 'middle':
                    extra /= 2.
                    add_top_padding(cell, extra)
                    cell.padding_bottom += extra
                else:
                    cell.padding_bottom += extra

            position_y += row.height + border_spacing_y

        group = group.copy_with_children(new_group_children)
        new_table_children.append(group)

        # Set missing baselines in a second loop because of rowspan
        for row in group.children:
            if row.baseline is None:
                if row.children:
                    # lowest bottom content edge
                    row.baseline = row.position_y - max(
                        cell.content_box_y() + cell.height
                        for cell in row.children)
                else:
                    row.baseline = 0
        group.height = position_y - group.position_y
        if group.children:
            # The last border spacing is outside of the group.
            group.height -= border_spacing_y
    table = table.copy_with_children(new_table_children)

    # If the height property has a bigger value, just add blank space
    # below the last row group.
    table.height = max(
        table.height if table.height != 'auto' else 0,
        position_y - table.content_box_y())

    # Layout for column groups and columns
    columns_height = position_y - initial_position_y
    if table.children:
        # The last border spacing is below the columns.
        columns_height -= border_spacing_y
    for group in table.column_groups:
        for column in group.children:
            resolve_percentages(column, containing_block=table)
            column.position_x = column_positions[column.grid_x]
            column.position_y = initial_position_y
            column.width = column_widths[column.grid_x]
            column.height = columns_height
        resolve_percentages(group, containing_block=table)
        first = group.children[0]
        last = group.children[-1]
        group.position_x = first.position_x
        group.position_y = initial_position_y
        group.width = last.position_x + last.width - first.position_x
        group.height = columns_height


    if ((table.position_y + table.margin_height()) > max_position_y
            and not page_is_empty):
        # If the table does not fit, put it on the next page.
        # (No page break inside tables yet.)
        table = None
    resume_at = None
    next_page = 'any'
    adjoining_margins = []
    collapsing_through = False
    return table, resume_at, next_page, adjoining_margins, collapsing_through


def add_top_padding(box, extra_padding):
    """Increase the top padding of a box. This also translates the children.
    """
    box.padding_top += extra_padding
    for child in box.children:
        child.translate(dy=extra_padding)


def fixed_table_layout(table):
    """Run the fixed table layout and return a list of column widths

    http://www.w3.org/TR/CSS21/tables.html#fixed-table-layout

    """
    assert table.width != 'auto'

    all_columns = [column for column_group in table.column_groups
                          for column in column_group.children]
    if table.children and table.children[0].children:
        first_rowgroup = table.children[0]
        first_row_cells = first_rowgroup.children[0].children
    else:
        first_row_cells = []
    num_columns = max(
        len(all_columns),
        sum(cell.colspan for cell in first_row_cells)
    )
    # ``None`` means not know yet.
    column_widths = [None] * num_columns

    # `width` on column boxes
    for i, column in enumerate(all_columns):
        resolve_one_percentage(column, 'width', table.width, ['auto'])
        if column.width != 'auto':
            column_widths[i] = column.width

    # `width` on cells of the first row.
    border_spacing_x, border_spacing_y = table.style.border_spacing
    i = 0
    for cell in first_row_cells:
        resolve_percentages(cell, table)
        if cell.width != 'auto':
            width = cell.border_width()
            width -= border_spacing_x * (cell.colspan - 1)
            # In the general case, this width affects several columns (through
            # colspan) some of which already have a width. Subscract these
            # known widths and divide among remaining columns.
            columns_without_width = []  # and occupied by this cell
            for j in xrange(i, i + cell.colspan):
                if column_widths[j] is None:
                    columns_without_width.append(j)
                else:
                    width -= column_widths[j]
            width_per_column = width / len(columns_without_width)
            for j in columns_without_width:
                column_widths[j] = width_per_column
        i += cell.colspan

    # Distribute the remaining space equally on columns that do not have
    # a width yet.
    all_border_spacing = border_spacing_x * (num_columns + 1)
    min_table_width = (sum(w for w in column_widths if w is not None)
                       + all_border_spacing)
    columns_without_width = [i for i, width in enumerate(column_widths)
                               if width is None]
    if columns_without_width and table.width >= min_table_width:
        remaining_width = table.width - min_table_width
        width_per_column = remaining_width / len(columns_without_width)
        for i in columns_without_width:
            column_widths[i] = width_per_column
    else:
        # XXX this is bad, but we were given a broken table to work with...
        for i in columns_without_width:
            column_widths[i] = 0

    # If the sum is less than the table width,
    # distribute the remaining space equally
    extra_width = table.width - sum(column_widths) - all_border_spacing
    if extra_width <= 0:
        # substract a negative: widen the table
        table.width -= extra_width
    elif num_columns:
        extra_per_column = extra_width / num_columns
        column_widths = [w + extra_per_column for w in column_widths]

    # Now we have table.width == sum(column_widths) + all_border_spacing
    # with possible floating point rounding errors.
    # (unless there is zero column)
    table.column_widths = column_widths
    return column_widths


def cell_baseline(cell):
    """
    Return the y position of a cell’s baseline from the top of its border box.

    See http://www.w3.org/TR/CSS21/tables.html#height-layout

    """
    # Do not use cell.descendants() as we do not want to recurse into
    # out-of-flow children.
    stack = [iter(cell.children)]  # DIY recursion
    while stack:
        child = next(stack[-1], None)
        if child is None:
            stack.pop()
            continue
        if not child.is_in_normal_flow():
            continue
        if isinstance(child, (boxes.LineBox, boxes.TableRowBox)):
            # First in-flow line or row.
            return child.baseline + child.position_y - cell.position_y
        if isinstance(child, boxes.ParentBox):
            stack.append(iter(child.children))
    # Default to the bottom of the content area.
    return cell.border_top_width + cell.padding_top + cell.height



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Handle various image formats.
"""

from __future__ import division, unicode_literals

from io import BytesIO
import contextlib

import cairo

from .utils import urlopen
from .css.computed_values import LENGTHS_TO_PIXELS
from .logger import LOGGER


# Map MIME types to functions that take a byte stream and return
# ``(pattern, width, height)`` a cairo Pattern and its dimension in pixels.
FORMAT_HANDLERS = {}

# TODO: currently CairoSVG only support images with an explicit
# width and height. When it supports images with only an intrinsic ratio
# this API will need to change.


def register_format(mime_type):
    """Register a handler for a give image MIME type."""
    def _decorator(function):
        FORMAT_HANDLERS[mime_type] = function
        return function
    return _decorator


@register_format('image/png')
def png_handler(file_like, _uri):
    """Return a cairo Surface from a PNG byte stream."""
    surface = cairo.ImageSurface.create_from_png(file_like)
    pattern = cairo.SurfacePattern(surface)
    return pattern, surface.get_width(), surface.get_height()


@register_format('image/svg+xml')
def cairosvg_handler(file_like, uri):
    """Return a cairo Surface from a SVG byte stream.

    This handler uses CairoSVG: http://cairosvg.org/
    """
    try:
        import cairosvg as _
    except ImportError as exception:
        return exception
    from cairosvg.surface import SVGSurface
    from cairosvg.parser import Tree, ParseError

    class ScaledSVGSurface(SVGSurface):
        """
        Have the cairo Surface object have intrinsic dimension
        in pixels instead of points.
        """
        @property
        def device_units_per_user_units(self):
            scale = super(ScaledSVGSurface, self).device_units_per_user_units
            return scale * LENGTHS_TO_PIXELS['pt']

    if uri.startswith('data:'):
        # Don’t pass data URIs to CairoSVG.
        # They are useless for relative URIs anyway.
        uri = None
    try:
        # Draw to a cairo surface but do not write to a file
        tree = Tree(file_obj=file_like, url=uri)
        surface = ScaledSVGSurface(tree, output=None, dpi=96)
    except (ParseError, NotImplementedError) as exception:
        return exception
    pattern = cairo.SurfacePattern(surface.cairo)
    return pattern, surface.width, surface.height


def fallback_handler(file_like, uri):
    """
    Parse a byte stream with PIL and return a cairo Surface.

    PIL supports many raster image formats and does not take a `format`
    parameter, it guesses the format from the content.
    """
    try:
        import pystacia as _
    except ImportError as exception:
        return exception
    from pystacia import read_blob
    from pystacia.util import TinyException
    try:
        with contextlib.closing(read_blob(file_like.read())) as image:
            png_bytes = image.get_blob('png')
    except TinyException as exception:
        return exception
    else:
        return png_handler(BytesIO(png_bytes), uri)


def get_image_from_uri(uri, type_=None):
    """Get a :class:`cairo.Surface`` from an image URI."""
    try:
        file_like, mime_type, _charset = urlopen(uri)
    except (IOError, ValueError) as exc:
        LOGGER.warn('Error while fetching an image from %s : %r', uri, exc)
        return None

    if not type_:
        type_ = mime_type  # Use eg. the HTTP header
    #else: the type was forced by eg. a 'type' attribute on <embed>
    handler = FORMAT_HANDLERS.get(type_, fallback_handler)
    exception = None
    try:
        image = handler(file_like, uri)
    except (IOError, MemoryError) as e:
        exception = e # Network or parsing error
    else:
        if isinstance(image, Exception):
            exception = image
    finally:
        try:
            file_like.close()
        except Exception:
            # Do not hide a more relevant exception.
            pass

    if exception is None:
        return image
    else:
        LOGGER.warn('Error while parsing an image at %s : %r', uri, exception)
        return None



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Test the public API.

"""

from __future__ import division, unicode_literals

import sys
import os
import io
import contextlib
import threading
import shutil
import tempfile

import pystacia

from .testing_utils import (
    resource_filename, assert_no_logs, TEST_UA_STYLESHEET)
from ..compat import urljoin
from .. import HTML, CSS
from .. import __main__


CHDIR_LOCK = threading.Lock()

@contextlib.contextmanager
def chdir(path):
    """Change the current directory in a context manager."""
    with CHDIR_LOCK:
        old_dir = os.getcwd()
        try:
            os.chdir(path)
            yield
        finally:
            os.chdir(old_dir)


@contextlib.contextmanager
def temp_directory():
    """Context manager that gives the path to a new temporary directory.

    Remove everything on exiting the context.

    """
    directory = tempfile.mkdtemp()
    try:
        yield directory
    finally:
        shutil.rmtree(directory)


def read_file(filename):
    """Shortcut for reading a file."""
    with open(filename, 'rb') as fd:
        return fd.read()


def write_file(filename, content):
    """Shortcut for reading a file."""
    with open(filename, 'wb') as fd:
        fd.write(content)


class TestHTML(HTML):
    """Like HTML, but with the testing (smaller) UA stylesheet"""
    def _ua_stylesheet(self):
        return [TEST_UA_STYLESHEET]


def _test_resource(class_, basename, check, **kwargs):
    """Common code for testing the HTML and CSS classes."""
    absolute_filename = resource_filename(basename)
    check(class_(absolute_filename, **kwargs))
    check(class_(guess=absolute_filename, **kwargs))
    check(class_(filename=absolute_filename, **kwargs))
    check(class_('file://' + absolute_filename, **kwargs))
    check(class_(guess='file://' + absolute_filename, **kwargs))
    check(class_(url='file://' + absolute_filename, **kwargs))
    with open(absolute_filename, 'rb') as fd:
        check(class_(fd, **kwargs))
    with open(absolute_filename, 'rb') as fd:
        check(class_(guess=fd, **kwargs))
    with open(absolute_filename, 'rb') as fd:
        check(class_(file_obj=fd, **kwargs))
    with open(absolute_filename, 'rb') as fd:
        content = fd.read()
    with chdir(os.path.dirname(__file__)):
        relative_filename = os.path.join('resources', basename)
        check(class_(relative_filename, **kwargs))
        check(class_(string=content, base_url=relative_filename, **kwargs))
        encoding = kwargs.get('encoding') or 'utf8'
        check(class_(string=content.decode(encoding),  # unicode
                        base_url=relative_filename, **kwargs))


@assert_no_logs
def test_html_parsing():
    """Test the constructor for the HTML class."""
    def check_doc1(html):
        """Check that a parsed HTML document looks like resources/doc1.html"""
        assert html.root_element.tag == 'html'
        assert [child.tag for child in html.root_element] == ['head', 'body']
        _head, body = html.root_element
        assert [child.tag for child in body] == ['h1', 'p', 'ul']
        h1 = body[0]
        assert h1.text == 'WeasyPrint test document (with Ünicōde)'
        url = urljoin(h1.base_url, 'pattern.png')
        assert url.startswith('file:')
        assert url.endswith('weasyprint/tests/resources/pattern.png')

    _test_resource(TestHTML, 'doc1.html', check_doc1)
    _test_resource(TestHTML, 'doc1-utf32.html', check_doc1, encoding='utf32')


@assert_no_logs
def test_css_parsing():
    """Test the constructor for the CSS class."""
    def check_css(css):
        """Check that a parsed stylsheet looks like resources/utf8-test.css"""
        # Using 'encoding' adds a CSSCharsetRule
        rule = css.stylesheet.cssRules[-1]
        assert rule.selectorText == 'h1::before'
        content, background = rule.style.getProperties(all=True)

        assert content.name == 'content'
        string, = content.propertyValue
        assert string.value == 'I løvë Unicode'

        assert background.name == 'background-image'
        url, = background.propertyValue
        url = url.absoluteUri
        assert url.startswith('file:')
        assert url.endswith('weasyprint/tests/resources/pattern.png')

    _test_resource(CSS, 'utf8-test.css', check_css)
    _test_resource(CSS, 'latin1-test.css', check_css, encoding='latin1')


def check_png_pattern(png_bytes):
    with contextlib.closing(pystacia.read_blob(png_bytes)) as image:
        assert image.size == (8, 8)
        lines = image.get_raw('rgba')['raw']
    from .test_draw import _, r, B, assert_pixels_equal
    assert_pixels_equal('api_png', 8, 8, lines, b''.join([
        _+_+_+_+_+_+_+_,
        _+_+_+_+_+_+_+_,
        _+_+r+B+B+B+_+_,
        _+_+B+B+B+B+_+_,
        _+_+B+B+B+B+_+_,
        _+_+B+B+B+B+_+_,
        _+_+_+_+_+_+_+_,
        _+_+_+_+_+_+_+_,
    ]))


@assert_no_logs
def test_python_render():
    """Test rendering with the Python API."""
    html = TestHTML(string='<body><img src=pattern.png>',
        base_url=resource_filename('dummy.html'))
    css = CSS(string='''
        @page { margin: 2px; -weasy-size: 8px; background: #fff }
        body { margin: 0; }
    ''')

    png_bytes = html.write_png(stylesheets=[css])
    pdf_bytes = html.write_pdf(stylesheets=[css])
    assert png_bytes.startswith(b'\211PNG\r\n\032\n')
    assert pdf_bytes.startswith(b'%PDF')

    check_png_pattern(png_bytes)
    # TODO: check PDF content? How?

    class fake_file(object):
        def __init__(self):
            self.chunks = []
            self.write = self.chunks.append

        def getvalue(self):
            return b''.join(self.chunks)
    png_file = fake_file()
    html.write_png(png_file, stylesheets=[css])
    assert png_file.getvalue() == png_bytes
    pdf_file = fake_file()
    html.write_pdf(pdf_file, stylesheets=[css])
    assert pdf_file.getvalue() == pdf_bytes

    with temp_directory() as temp:
        png_filename = os.path.join(temp, '1.png')
        pdf_filename = os.path.join(temp, '1.pdf')
        html.write_png(png_filename, stylesheets=[css])
        html.write_pdf(pdf_filename, stylesheets=[css])
        assert read_file(png_filename) == png_bytes
        assert read_file(pdf_filename) == pdf_bytes

        png_filename = os.path.join(temp, '2.png')
        pdf_filename = os.path.join(temp, '2.pdf')
        with open(png_filename, 'wb') as png_file:
            html.write_png(png_file, stylesheets=[css])
        with open(pdf_filename, 'wb') as pdf_file:
            html.write_pdf(pdf_file, stylesheets=[css])
        assert read_file(png_filename) == png_bytes
        assert read_file(pdf_filename) == pdf_bytes


@assert_no_logs
def test_command_line_render():
    """Test rendering with the command-line API."""
    css = b'''
        @page { margin: 2px; -weasy-size: 8px; background: #fff }
        body { margin: 0; }
    '''
    html = b'<body><img src=pattern.png>'
    combined = b'<style>' + css + b'</style>' + html
    linked = b'<link rel=stylesheet href=style.css>' + html

    with chdir(resource_filename('')):
        # Reference
        png_bytes = TestHTML(string=combined, base_url='dummy.html').write_png()
        pdf_bytes = TestHTML(string=combined, base_url='dummy.html').write_pdf()
    check_png_pattern(png_bytes)

    def run(args, stdin=b''):
        stdin = io.BytesIO(stdin)
        stdout = io.BytesIO()
        try:
            __main__.HTML = TestHTML
            __main__.main(args.split(), stdin=stdin, stdout=stdout)
        finally:
            __main__.HTML = HTML
        return stdout.getvalue()

    with temp_directory() as temp:
        with chdir(temp):
            pattern_bytes = read_file(resource_filename('pattern.png'))
            write_file('pattern.png', pattern_bytes)
            write_file('no_css.html', html)
            write_file('combined.html', combined)
            write_file('combined-utf32.html',
                combined.decode('ascii').encode('utf32'))
            write_file('linked.html', linked)
            write_file('style.css', css)

            run('combined.html out1.png')
            run('combined.html out2.pdf')
            assert read_file('out1.png') == png_bytes
            assert read_file('out2.pdf') == pdf_bytes

            run('combined-utf32.html out3.png --encoding utf32')
            assert read_file('out3.png') == png_bytes

            combined_absolute = os.path.join(temp, 'combined.html')
            run(combined_absolute + ' out4.png')
            assert read_file('out4.png') == png_bytes

            combined_url = 'file://{0}/{1}'.format(temp, 'combined.html')
            run(combined_url + ' out5.png')
            assert read_file('out5.png') == png_bytes

            run('linked.html out6.png')  # test relative URLs
            assert read_file('out6.png') == png_bytes

            run('combined.html out7 -f png')
            run('combined.html out8 --format pdf')
            assert read_file('out7') == png_bytes
            assert read_file('out8') == pdf_bytes

            run('no_css.html out9.png')
            run('no_css.html out10.png -s style.css')
            assert read_file('out9.png') != png_bytes
            assert read_file('out10.png') == png_bytes

            stdout = run('--format png combined.html -')
            assert stdout == png_bytes

            run('- out11.png', stdin=combined)
            check_png_pattern(read_file('out11.png'))
            assert read_file('out11.png') == png_bytes

            stdout = run('--format png - -', stdin=combined)
            assert stdout == png_bytes



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Test the management of the CSS properties.

"""

from __future__ import division, unicode_literals

from cssutils.css import PropertyValue
from pytest import raises

from .testing_utils import assert_no_logs
from ..css import validation
from ..css.values import as_css


def expand_to_dict(short_name, short_values):
    """Helper to test shorthand properties expander functions."""
    return dict((name, as_css(value) if isinstance(value, (list, tuple))
                       else getattr(value, 'cssText', value))
                for name, value in validation.EXPANDERS[short_name](
                    short_name, list(PropertyValue(short_values))))


@assert_no_logs
def test_expand_four_sides():
    """Test the 4-value properties."""
    assert expand_to_dict('margin', 'inherit') == {
        'margin_top': 'inherit',
        'margin_right': 'inherit',
        'margin_bottom': 'inherit',
        'margin_left': 'inherit',
    }
    assert expand_to_dict('margin', '1em') == {
        'margin_top': '1em',
        'margin_right': '1em',
        'margin_bottom': '1em',
        'margin_left': '1em',
    }
    assert expand_to_dict('padding', '1em 0') == {
        'padding_top': '1em',
        'padding_right': '0',
        'padding_bottom': '1em',
        'padding_left': '0',
    }
    assert expand_to_dict('padding', '1em 0 2em') == {
        'padding_top': '1em',
        'padding_right': '0',
        'padding_bottom': '2em',
        'padding_left': '0',
    }
    assert expand_to_dict('padding', '1em 0 2em 5px') == {
        'padding_top': '1em',
        'padding_right': '0',
        'padding_bottom': '2em',
        'padding_left': '5px',
    }
    with raises(ValueError):
        expand_to_dict('padding', '1 2 3 4 5')


@assert_no_logs
def test_expand_borders():
    """Test the ``border`` property."""
    assert expand_to_dict('border_top', '3px dotted red') == {
        'border_top_width': '3px',
        'border_top_style': 'dotted',
        'border_top_color': 'red',
    }
    assert expand_to_dict('border_top', '3px dotted') == {
        'border_top_width': '3px',
        'border_top_style': 'dotted',
        'border_top_color': 'currentColor',
    }
    assert expand_to_dict('border_top', '3px red') == {
        'border_top_width': '3px',
        'border_top_style': 'none',
        'border_top_color': 'red',
    }
    assert expand_to_dict('border_top', 'solid') == {
        'border_top_width': 3,
        'border_top_style': 'solid',
        'border_top_color': 'currentColor',
    }
    assert expand_to_dict('border', '6px dashed green') == {
        'border_top_width': '6px',
        'border_top_style': 'dashed',
        'border_top_color': 'green',

        'border_left_width': '6px',
        'border_left_style': 'dashed',
        'border_left_color': 'green',

        'border_bottom_width': '6px',
        'border_bottom_style': 'dashed',
        'border_bottom_color': 'green',

        'border_right_width': '6px',
        'border_right_style': 'dashed',
        'border_right_color': 'green',
    }
    with raises(ValueError):
        expand_to_dict('border', '6px dashed left')


@assert_no_logs
def test_expand_list_style():
    """Test the ``list_style`` property."""
    assert expand_to_dict('list_style', 'inherit') == {
        'list_style_position': 'inherit',
        'list_style_image': 'inherit',
        'list_style_type': 'inherit',
    }
    assert expand_to_dict('list_style', 'url(foo.png)') == {
        'list_style_position': 'outside',
        'list_style_image': 'foo.png',
        'list_style_type': 'disc',
    }
    assert expand_to_dict('list_style', 'square') == {
        'list_style_position': 'outside',
        'list_style_image': 'none',
        'list_style_type': 'square',
    }
    assert expand_to_dict('list_style', 'circle inside') == {
        'list_style_position': 'inside',
        'list_style_image': 'none',
        'list_style_type': 'circle',
    }
    with raises(ValueError):
        expand_to_dict('list_style', 'red')


def assert_background(css, **kwargs):
    """Helper checking the background properties."""
    expanded = expand_to_dict('background', css).items()
    expected = [('background_' + key, value)
                for key, value in kwargs.items()]
    assert sorted(expanded) == sorted(expected)


@assert_no_logs
def test_expand_background():
    """Test the ``background`` property."""
    assert_background(
        'red',
        color='red', ##
        image='none',
        repeat='repeat',
        attachment='scroll',
        position='0% 0%'

    )
    assert_background(
        'url(foo.png)',
        color='transparent',
        image='foo.png', ##
        repeat='repeat',
        attachment='scroll',
        position='0% 0%'
    )
    assert_background(
        'no-repeat',
        color='transparent',
        image='none',
        repeat='no-repeat', ##
        attachment='scroll',
        position='0% 0%'
    )
    assert_background(
        'fixed',
        color='transparent',
        image='none',
        repeat='repeat',
        attachment='fixed', ##
        position='0% 0%'
    )
    assert_background(
        'top right',
        color='transparent',
        image='none',
        repeat='repeat',
        attachment='scroll',
        # Order swapped to be in (horizontal, vertical) order.
        position='100% 0%' ##
    )
    assert_background(
        'url(bar) #f00 repeat-y center left fixed',
        color='#f00', ##
        image='bar', ##
        repeat='repeat-y', ##
        attachment='fixed', ##
        # Order swapped to be in (horizontal, vertical) order.
        position='0% 50%' ##
    )
    assert_background(
        '#00f 10% 200px',
        color='#00f', ##
        image='none',
        repeat='repeat',
        attachment='scroll',
        position='10% 200px' ##
    )
    assert_background(
        'right 78px fixed',
        color='transparent',
        image='none',
        repeat='repeat',
        attachment='fixed', ##
        position='100% 78px' ##
    )


@assert_no_logs
def test_font():
    """Test the ``font`` property."""
    assert expand_to_dict('font', '12px sans_serif') == {
        'font_style': 'normal',
        'font_variant': 'normal',
        'font_weight': 400,
        'font_size': '12px', ##
        'line_height': 'normal',
        'font_family': 'sans_serif', ##
    }
    assert expand_to_dict('font', 'small/1.2 "Some Font", serif') == {
        'font_style': 'normal',
        'font_variant': 'normal',
        'font_weight': 400,
        'font_size': 'small', ##
        'line_height': '1.2', ##
        # The comma and quotes were lost in expand_to_dict()
        'font_family': 'Some Font serif', ##
    }
    assert expand_to_dict('font', 'small-caps italic 700 large serif') == {
        'font_style': 'italic', ##
        'font_variant': 'small-caps', ##
        'font_weight': 700, ##
        'font_size': 'large', ##
        'line_height': 'normal',
        'font_family': 'serif', ##
    }



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
WeasyPrint testing suite.

"""

from __future__ import division, unicode_literals, print_function

import sys
import os.path
import logging
import contextlib
import functools

from .. import HTML
from ..document import PNGDocument
from ..css import PARSER as CSS_PARSER
from ..logger import LOGGER

# TODO: find a way to not depend on a specific font
FONTS = 'Liberation Sans, Arial'

TEST_UA_STYLESHEET = CSS_PARSER.parseFile(os.path.join(
    os.path.dirname(__file__), '..', 'css', 'tests_ua.css'
))


class TestPNGDocument(PNGDocument):
    """Like PNGDocument, but with a different user-agent stylesheet.

    This stylesheet is shorter, which makes tests faster.

    """
    def __init__(self, html_source, base_url=None):
        super(TestPNGDocument, self).__init__(
            HTML(string=html_source, base_url=base_url).root_element,
            user_stylesheets=[],
            user_agent_stylesheets=[TEST_UA_STYLESHEET])


def resource_filename(basename):
    """Return the absolute path of the resource called ``basename``."""
    return os.path.join(os.path.dirname(__file__), 'resources', basename)


class CallbackHandler(logging.Handler):
    """A logging handler that calls a function for every message."""
    def __init__(self, callback):
        logging.Handler.__init__(self)
        self.emit = callback


@contextlib.contextmanager
def capture_logs():
    """Return a context manager that captures all logged messages."""
    loggers = [LOGGER, logging.getLogger('CSSUTILS')]
    previous_handlers = []
    messages = []

    def emit(record):
        message = '%s: %s' % (record.levelname.upper(), record.getMessage())
        messages.append(message)
        print(message, file=sys.stderr)

    for logger in loggers:
        previous_handlers.append((logger, logger.handlers))
        logger.handlers = []
        logger.addHandler(CallbackHandler(emit))
    try:
        yield messages
    finally:
        for logger, handlers in previous_handlers:
            logger.handlers = handlers


def assert_no_logs(function):
    """Decorator that asserts that nothing is logged in a function."""
    @functools.wraps(function)
    def wrapper():
        with capture_logs() as logs:
            try:
                function()
            except Exception:
                if logs:
                    print('%i errors logged:' % len(logs), file=sys.stderr)
                raise
            else:
                assert len(logs) == 0, '%i errors logged:' % len(logs)
    return wrapper



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Test the base mechanisms of CSS.

"""

from __future__ import division, unicode_literals

import os.path

import cssutils
from pytest import raises

from .testing_utils import (
    resource_filename, assert_no_logs, capture_logs,
    TestPNGDocument, TEST_UA_STYLESHEET)
from .. import css
from ..css.computed_values import used_line_height
from ..document import PNGDocument
from .. import HTML


def parse_html(filename, **kwargs):
    """Parse an HTML file from the test resources and resolve relative URL."""
    html = HTML(filename=resource_filename(filename))
    kwargs.setdefault('user_agent_stylesheets', [TEST_UA_STYLESHEET])
    kwargs.setdefault('user_stylesheets', [])
    return PNGDocument(html.root_element, **kwargs)


@assert_no_logs
def test_style_dict():
    """Test a style in a ``dict``."""
    style = css.StyleDict({
        'margin_left': 12,
        'display': 'block'})
    assert style.display == 'block'
    assert style.margin_left == 12
    with raises(KeyError):
        style.position  # pylint: disable=W0104


@assert_no_logs
def test_find_stylesheets():
    """Test if the stylesheets are found in a HTML document."""
    document = parse_html('doc1.html')

    sheets = list(css.find_stylesheets(document))
    assert len(sheets) == 3
    # Also test that stylesheets are in tree order
    assert [s.href.rsplit('/', 1)[-1].rsplit(',', 1)[-1] for s in sheets] \
        == ['sheet1.css', 'a%7Bcolor%3AcurrentColor%7D',
            'doc1.html']

    rules = list(rule for sheet in sheets
                      for rule in css.effective_rules(sheet, 'print'))
    assert len(rules) == 10
    # Also test appearance order
    assert [rule.selectorText for rule in rules] \
        == ['a', 'li', 'p', 'ul', 'li', 'a:after', ':first', 'ul',
            'body > h1:first-child', 'h1 ~ p ~ ul a:after']


@assert_no_logs
def test_expand_shorthands():
    """Test the expand shorthands."""
    sheet = cssutils.parseFile(resource_filename('sheet2.css'))
    assert sheet.cssRules[0].selectorText == 'li'

    style = sheet.cssRules[0].style
    assert style['margin'] == '2em 0'
    assert style['margin-bottom'] == '3em'
    assert style['margin-left'] == '4em'
    assert not style['margin-top']

    style = dict(
        (name, css.values.as_css([value]))
        for name, (value, _priority) in css.effective_declarations(style))

    assert 'margin' not in style
    assert style['margin_top'] == '2em'
    assert style['margin_right'] == '0'
    assert style['margin_bottom'] == '2em', \
        '3em was before the shorthand, should be masked'
    assert style['margin_left'] == '4em', \
        '4em was after the shorthand, should not be masked'


def parse_css(filename):
    """Parse and return the CSS at ``filename``."""
    return cssutils.parseFile(resource_filename(filename))


@assert_no_logs
def test_annotate_document():
    """Test a document with inline style."""
    # Short names for variables are OK here
    # pylint: disable=C0103

    document = parse_html(
        'doc1.html',
        user_stylesheets=[parse_css('user.css')],
        user_agent_stylesheets=[parse_css('mini_ua.css')],
    )

    # Element objects behave a lists of their children
    _head, body = document.dom
    h1, p, ul = body
    li_0, _li_1 = ul
    a, = li_0

    h1 = document.style_for(h1)
    p = document.style_for(p)
    ul = document.style_for(ul)
    li_0 = document.style_for(li_0)
    after = document.style_for(a, 'after')
    a = document.style_for(a)

    assert h1.background_image == 'file://' \
        + os.path.abspath(resource_filename('logo_small.png'))

    assert h1.font_weight == 700

    # 32px = 1em * font-size: 2em * initial 16px
    assert p.margin_top == 32
    assert p.margin_right == 0
    assert p.margin_bottom == 32
    assert p.margin_left == 0

    # 32px = 2em * initial 16px
    assert ul.margin_top == 32
    assert ul.margin_right == 32
    assert ul.margin_bottom == 32
    assert ul.margin_left == 32

    # thick = 5px, 0.25 inches = 96*.25 = 24px
    assert ul.border_top_width == 0
    assert ul.border_right_width == 5
    assert ul.border_bottom_width == 0
    assert ul.border_left_width == 24

    # 32px = 2em * initial 16px
    # 64px = 4em * initial 16px
    assert li_0.margin_top == 32
    assert li_0.margin_right == 0
    assert li_0.margin_bottom == 32
    assert li_0.margin_left == 64

    assert a.text_decoration == frozenset(['underline'])

    assert a.padding_top == 1
    assert a.padding_right == 2
    assert a.padding_bottom == 3
    assert a.padding_left == 4

    color = a.color
    assert (color.red, color.green, color.blue, color.alpha) == (255, 0, 0, 1)
    # Test the initial border-color: currentColor
    color = a.border_top_color
    assert (color.red, color.green, color.blue, color.alpha) == (255, 0, 0, 1)

    # The href attr should be as in the source, not made absolute.
    assert after.content == [
        ('STRING', ' ['), ('STRING', 'home.html'), ('STRING', ']')]
    color = after.background_color
    assert (color.red, color.green, color.blue, color.alpha) == (255, 0, 0, 1)

    # TODO much more tests here: test that origin and selector precedence
    # and inheritance are correct, ...

    # pylint: enable=C0103


@assert_no_logs
def test_default_stylesheet():
    """Test if the user-agent stylesheet is used and applied."""
    document = parse_html('doc1.html')
    head_style = document.style_for(document.dom.head)
    assert head_style.display == 'none', \
        'The HTML4 user-agent stylesheet was not applied'


@assert_no_logs
def test_page():
    """Test the ``@page`` properties."""
    document = parse_html('doc1.html', user_stylesheets=[
        cssutils.parseString('''
            html {
                color: red;
            }
            @page {
                margin: 10px;
            }
            @page :right {
                color: blue;
                margin-bottom: 12pt;
                font-size: 20px;
                @top-left {
                    width: 10em;
                }
                @top-right {
                    font-size: 10px;
                }
            }
        ''')
    ])

    style = document.style_for('first_left_page')
    assert style.margin_top == 5
    assert style.margin_left == 10
    assert style.margin_bottom == 10
    assert style.color.cssText == 'red'  # inherited from html

    style = document.style_for('first_right_page')
    assert style.margin_top == 5
    assert style.margin_left == 10
    assert style.margin_bottom == 16
    assert style.color.cssText == 'blue'

    style = document.style_for('left_page')
    assert style.margin_top == 10
    assert style.margin_left == 10
    assert style.margin_bottom == 10
    assert style.color.cssText == 'red'  # inherited from html

    style = document.style_for('right_page')
    assert style.margin_top == 10
    assert style.margin_left == 10
    assert style.margin_bottom == 16
    assert style.color.cssText == 'blue'

    style = document.style_for('first_left_page', '@top-left')
    assert style is None

    style = document.style_for('first_right_page', '@top-left')
    assert style.font_size == 20  # inherited from @page
    assert style.width == 200

    style = document.style_for('first_right_page', '@top-right')
    assert style.font_size == 10


@assert_no_logs
def test_warnings():
    """Check that appropriate warnings are logged."""
    for source, messages in [
        ('<style>:link { margin: 2cm',
            ['WARNING: Unsupported selector']),
        ('<style>@page foo { margin: 2cm',
            ['WARNING: Unsupported @page selector']),
        ('<link rel=stylesheet href=data:image/png,>',
            ['WARNING: Expected `text/css` for stylsheet at']),
        ('<style>foo { margin-color: red',
            ['WARNING: Ignored declaration', 'unknown property']),
        ('<style>foo { margin-top: red',
            ['WARNING: Ignored declaration', 'invalid value']),
        ('<html style="margin-color: red">',
            ['WARNING: Ignored declaration', 'unknown property']),
        ('<html style="margin-top: red">',
            ['WARNING: Ignored declaration', 'invalid value']),
    ]:
        with capture_logs() as logs:
            TestPNGDocument(source).style_for('')
        assert len(logs) == 1
        for message in messages:
            assert message in logs[0]


@assert_no_logs
def test_error_recovery():
    with capture_logs() as logs:
        document = TestPNGDocument('''
            <style> html { color red; color: blue; color
        ''')
        html = document.formatting_structure
        assert html.style.color.value == 'blue'

        document = TestPNGDocument('''
            <html style="color; color: blue; color red">
        ''')
        html = document.formatting_structure
        assert html.style.color.value == 'blue'
    assert len(logs) == 12


@assert_no_logs
def test_line_height_inheritance():
    document = TestPNGDocument('''
        <style>
            html { font-size: 10px; line-height: 140% }
            section { font-size: 10px; line-height: 1.4 }
            div, p { font-size: 20px; vertical-align: 50% }
        </style>
        <body><div><section><p></p></section></div></body>
    ''')
    html = document.formatting_structure
    body, = html.children
    div, = body.children
    section, = div.children
    paragraph, = section.children
    assert html.style.font_size == 10
    assert div.style.font_size == 20
    # 140% of 10px = 14px is inherited from html
    assert used_line_height(div.style) == 14
    assert div.style.vertical_align == 7  # 50 % of 14px

    assert paragraph.style.font_size == 20
    # 1.4 is inherited from p, 1.4 * 20px on em = 28px
    assert used_line_height(paragraph.style) == 28
    assert paragraph.style.vertical_align == 14  # 50% of 28pxhh



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Functions resolving percentages.

"""

from __future__ import division, unicode_literals

from ..formatting_structure import boxes
from ..css.values import get_percentage_value


def resolve_one_percentage(box, property_name, refer_to,
                           allowed_keywords=None):
    """Set a used length value from a computed length value.

    ``refer_to`` is the length for 100%. If ``refer_to`` is not a number, it
    just replaces percentages.

    """
    # box.style has computed values
    values = box.style[property_name]
    if isinstance(values, (int, float)):
        # Absolute length (was converted to pixels in "computed values")
        result = values
    else:
        percentage = get_percentage_value(values)
        if percentage is not None:
            if isinstance(refer_to, (int, float)):
                # A percentage
                result = percentage * refer_to / 100.
            else:
                # Replace percentages when we have no refer_to that
                # makes sense.
                result = refer_to
        else:
            # Some other values such as 'auto' may be allowed
            result = values
            assert allowed_keywords and result in allowed_keywords
    # box attributes are used values
    setattr(box, property_name, result)


def resolve_percentages(box, containing_block):
    """Set used values as attributes of the box object."""
    if isinstance(containing_block, boxes.Box):
        # cb is short for containing block
        cb_width = containing_block.width
        cb_height = containing_block.height
    else:
        cb_width, cb_height = containing_block
    if isinstance(box, boxes.PageBox):
        maybe_height = cb_height
    else:
        maybe_height = cb_width
    resolve_one_percentage(box, 'margin_left', cb_width, ['auto'])
    resolve_one_percentage(box, 'margin_right', cb_width, ['auto'])
    resolve_one_percentage(box, 'margin_top', maybe_height, ['auto'])
    resolve_one_percentage(box, 'margin_bottom', maybe_height, ['auto'])
    resolve_one_percentage(box, 'padding_left', cb_width)
    resolve_one_percentage(box, 'padding_right', cb_width)
    resolve_one_percentage(box, 'padding_top', maybe_height)
    resolve_one_percentage(box, 'padding_bottom', maybe_height)
    resolve_one_percentage(box, 'width', cb_width, ['auto'])
    # Not supported yet:
#    resolve_one_percentage(box, 'min_width', cb_width)
#    resolve_one_percentage(box, 'max_width', cb_width, ['none'])

    # XXX later: top, bottom, left and right on positioned elements

    if cb_height == 'auto':
        # Special handling when the height of the containing block
        # depends on its content.
        resolve_one_percentage(box, 'height', 'auto', ['auto'])
        # Not supported yet, but min_height is used for margin collapsing.
        resolve_one_percentage(box, 'min_height', 0)
#        resolve_one_percentage(box, 'max_height', None, ['none'])
    else:
        resolve_one_percentage(box, 'height', cb_height, ['auto'])
        # Not supported yet, but min_height is used for margin collapsing.
        resolve_one_percentage(box, 'min_height', cb_height)
#        resolve_one_percentage(box, 'max_height', cb_height, ['none'])

    # Used value == computed value
    for side in ['top', 'right', 'bottom', 'left']:
        prop = 'border_{0}_width'.format(side)
        setattr(box, prop, box.style[prop])

    if box.style.box_sizing == 'border-box':
        if box.width != 'auto':
            box.width -= (box.padding_left + box.padding_right +
                          box.border_left_width + box.border_right_width)
        if box.height != 'auto':
            box.height -= (box.padding_top + box.padding_bottom +
                           box.border_top_width + box.border_bottom_width)
    else:
        assert box.style.box_sizing == 'content-box'



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
WeasyPrint
==========

WeasyPrint converts web documents, mainly HTML documents with CSS, to PDF.

"""

from __future__ import division, unicode_literals

# Make sure the logger is configured early:
from .logger import LOGGER


VERSION = '0.6'
__version__ = VERSION


# No import here. For this module, do them in functions/methods instead.
# (This reduces the work for eg. 'weasyprint --help')


class Resource(object):
    """Common API for creating instances of :class:`HTML` or :class:`CSS`.

    You can just create an instance with a positional argument
    (ie. ``HTML(something)``) and it will try to guess if the input is
    a filename, an absolute URL, or a file-like object.

    Alternatively, you can name the argument so that no guessing is
    involved:

    * ``HTML(filename=foo)`` a filename, absolute or relative to
      the current directory.
    * ``HTML(url=foo)`` an absolute, fully qualified URL.
    * ``HTML(file_obj=foo)`` a file-like object: any object with
      a :meth:`read` method.
    * ``HTML(string=foo)`` a string of HTML source.
      (This argument must be named.)

    Specifying multiple inputs is an error: ``HTML(filename=foo, url=bar)``

    Optional, additional named arguments:

    * ``encoding``: force the character encoding
    * ``base_url``: used to resolve relative URLs. If not passed explicitly,
      try to use the input filename, URL, or ``name`` attribute of
      file objects.

    """


class HTML(Resource):
    """Fetch and parse an HTML document with lxml.

    See :class:`Resource` to create an instance.

    """
    def __init__(self, guess=None, filename=None, url=None, file_obj=None,
                 string=None, encoding=None, base_url=None):
        import lxml.html
        from .utils import ensure_url

        source_type, source, base_url = _select_source(
            guess, filename, url, file_obj, string, base_url)

        if source_type == 'string':
            parse = lxml.html.document_fromstring
        else:
            parse = lxml.html.parse
            if source_type != 'file_obj':
                # If base_url is None we want the used base URL to be
                # an URL, not a filename.
                source = ensure_url(source)
        parser = lxml.html.HTMLParser(encoding=encoding)
        result = parse(source, base_url=base_url, parser=parser)
        if source_type != 'string':
            result = result.getroot()
        if result is None:
            raise ValueError('Error while parsing HTML')
        self.root_element = result

    def _ua_stylesheet(self):
        from .css import HTML5_UA_STYLESHEET
        return [HTML5_UA_STYLESHEET]

    def _write(self, document_class, target, stylesheets):
        return document_class(
            self.root_element,
            user_stylesheets=list(_parse_stylesheets(stylesheets)),
            user_agent_stylesheets=self._ua_stylesheet(),
        ).write_to(target)

    def write_pdf(self, target=None, stylesheets=None):
        """Render the document to PDF.

        :param target:
            a filename, file-like object, or :obj:`None`.
        :param stylesheets:
            a list of user stylsheets, as :class:`CSS` objects, filenames,
            URLs, or file-like objects
        :returns:
            If :obj:`target` is :obj:`None`, a PDF byte string.
        """
        from .document import PDFDocument
        return self._write(PDFDocument, target, stylesheets)

    def write_png(self, target=None, stylesheets=None):
        """Render the document to PNG.

        :param target:
            a filename, file-like object, or :obj:`None`.
        :param stylesheets:
            a list of user stylsheets, as :class:`CSS` objects, filenames,
            URLs, or file-like objects
        :returns:
            If :obj:`target` is :obj:`None`, a PNG byte string.
        """
        from .document import PNGDocument
        return self._write(PNGDocument, target, stylesheets)


class CSS(Resource):
    """Fetch and parse a CSS stylesheet with cssutils.

    See :class:`Resource` to create an instance. A :class:`CSS` object
    is not useful on its own but can be passed to :meth:`HTML.write_pdf` or
    :meth:`HTML.write_png`.

    """
    def __init__(self, guess=None, filename=None, url=None, file_obj=None,
                 string=None, encoding=None, base_url=None):
        from .css import PARSER

        source_type, source, base_url = _select_source(
            guess, filename, url, file_obj, string, base_url)

        if source_type == 'file_obj':
            source = source.read()
            source_type = 'string'
        if source_type == 'url':
            self.stylesheet = PARSER.parseUrl(source, encoding=encoding)
            if base_url is not None:
                # source and href are the same for parseUrl
                self.stylesheet.href = base_url
        else:
            parser = {'filename': PARSER.parseFile,
                      'string': PARSER.parseString}[source_type]
            self.stylesheet = parser(source, encoding=encoding, href=base_url)


def _select_source(guess, filename, url, file_obj, string, base_url):
    """
    Check that only one input is not None, and return it with the
    normalized ``base_url``.
    """
    from .utils import ensure_url
    from .compat import urlparse

    if base_url is not None:
        base_url = ensure_url(base_url)

    nones = [guess is None, filename is None, url is None,
             file_obj is None, string is None]
    if nones == [False, True, True, True, True]:
        if hasattr(guess, 'read'):
            type_ = 'file_obj'
            if base_url is None:
                # filesystem file objects have a 'name' attribute.
                name = getattr(guess, 'name', None)
                if name:
                    base_url = ensure_url(name)
        elif urlparse(guess).scheme:
            type_ = 'url'
        else:
            type_ = 'filename'
        return type_, guess, base_url
    if nones == [True, False, True, True, True]:
        return 'filename', filename, base_url
    if nones == [True, True, False, True, True]:
        return 'url', url, base_url
    if nones == [True, True, True, False, True]:
        if base_url is None:
            # filesystem file objects have a 'name' attribute.
            name = getattr(file_obj, 'name', None)
            if name:
                base_url = ensure_url(name)
        return 'file_obj', file_obj, base_url
    if nones == [True, True, True, True, False]:
        return 'string', string, base_url

    raise TypeError('Expected only one source, got %i' % nones.count(False))


def _parse_stylesheets(stylesheets):
    """Yield parsed cssutils stylesheets.

    Accept :obj:`None` or a list of filenames, urls or CSS objects.

    """
    if stylesheets is None:
        return
    for css in stylesheets:
        if hasattr(css, 'stylesheet'):
            yield css.stylesheet
        else:
            yield CSS(css).stylesheet



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Module managing the layout creation before drawing a document.

"""

from __future__ import division, unicode_literals

from .pages import make_all_pages, add_margin_boxes


def layout_document(document, root_box):
    """Lay out the whole document.

    This includes line breaks, page breaks, absolute size and position for all
    boxes.

    :param document: a Document object.
    :returns: a list of laid out Page objects.

    """
    pages = list(make_all_pages(document, root_box))
    return list(add_margin_boxes(document, pages))



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Margin box variable dimension computation.

http://dev.w3.org/csswg/css3-page/#margin-dimension

"""

from __future__ import division, unicode_literals


def with_rule_2(side_boxes, outer_sum):
    """Margin box variable dimension computation, in the case where
    rule 2 applies.
    """
    box_a, box_b, box_c = side_boxes
    seen = set()
    while 1:
        implementation, swap = IMPLEMENTATIONS[
            tuple(box.inner == 'auto' for box in side_boxes),
            tuple('auto' in [box.margin_a, box.margin_b] for box in side_boxes),
        ]
        assert implementation not in seen  # break loops
        seen.add(implementation)
        if swap:
            box_a, box_c = swap_ac(box_a, box_c)
        result = implementation(box_a, box_b, box_c, outer_sum)
        if swap:
            box_a, box_c = swap_ac(box_a, box_c)

        if result == 'ok':
            return

        # XXX
        if result is NotImplemented:
            for box in side_boxes:
                if box.margin_a == 'auto':
                    box.margin_a = 0
                if box.margin_b == 'auto':
                    box.margin_b = 0

        # Try again with less constraints
        box_a.margin_b = 'auto'
        box_c.margin_a = 'auto'


def swap_ac(box_a, box_c):
    """Swap margins A and B in each box, and return ``box_c, box_a``."""
    # TODO: find a better way to do this?
    for box in [box_a, box_c]:
        box.margin_a, box.margin_b = box.margin_b, box.margin_a
    return box_c, box_a


IMPLEMENTATIONS = {}


def register(auto_inners, auto_margins):
    """Register an implementation for with_rule_2()

    :param auto_inners:
        A tuple of 3 booleans. True if box A, B and C (respectively) have a
        value of 'auto' for the inner dimension.
    :param auto_margins:
        A tuple of 3 booleans. True if the box has any margin with
        the value 'auto'.

    """
    def decorator(function):
        key_1 = auto_inners, auto_margins
        keys = [(key_1, False)]
        # symmetry:
        key_2 = auto_inners[::-1], auto_margins[::-1]
        if key_2 != key_1:
            keys.append((key_2, True))
        for key, swap_ac in keys:
            assert key not in IMPLEMENTATIONS, (
                'redundant implementations for %r: %r' % (key, function))
            IMPLEMENTATIONS[key] = function, swap_ac
        return function
    return decorator


def outer(box, ignore_auto=''):
    """Return the outer dimension of a box, or raise if any value is auto."""
    result = box.padding_plus_border
    for value, type_ in [
        (box.inner, 'I'),
        (box.margin_a, 'M'),
        (box.margin_b, 'M'),
    ]:
        if value == 'auto':
            if type_ not in ignore_auto:
                assert False
        else:
            result += value
    return result


def distribute_margins(box, new_outer):
    """Set 'auto' margins on ``box`` so that the outer dimension
    is ``new_outer``.

    """
    num_auto = [box.margin_a, box.margin_b].count('auto')
    # Raises if num_auto == 0:
    each_auto = (new_outer - outer(box, ignore_auto='M')) / num_auto
    if box.margin_a == 'auto':
        box.margin_a = each_auto
    if box.margin_b == 'auto':
        box.margin_b = each_auto


def distribute_margins_and_inner(box, new_outer):
    """Set 'auto' values when both inner and some margins are 'auto'."""
    remaining = new_outer - outer(box, ignore_auto='MI')
    set_inner_within_range(box, remaining)
    distribute_margins(box, new_outer)


def set_inner_within_range(box, optimal_inner):
    """Set ``box.inner`` as close to ``optimal_inner`` as possible with
    range of its intrinsic bounds.

    """
    min_inner = box.minimum
    if optimal_inner < min_inner:
        box.inner = min_inner
    else:
        max_inner = box.preferred
        if optimal_inner > max_inner:
            box.inner = max_inner
        else:
            box.inner = optimal_inner


@register(auto_inners=(0, 0, 0), auto_margins=(0, 0, 0))
def implementation_1(box_a, box_b, box_c, outer_sum):
    # If the previous values matched the constraints,
    # they will end up unchanged
    box_a.margin_b = 'auto'
    box_c.margin_a = 'auto'
    target_outer_ac = (outer_sum - outer(box_b)) / 2
    distribute_margins(box_a, target_outer_ac)
    distribute_margins(box_c, target_outer_ac)
    return 'ok'


@register(auto_inners=(0, 0, 1), auto_margins=(0, 0, 0))
def implementation_2(box_a, box_b, box_c, outer_sum):
    outer_a = outer(box_a)
    outer_b = outer(box_b)
    if outer_a == (outer_sum - outer_b) / 2: # Rules 1 and 2
        outer_c = outer_a  # Rule 2
        inner_c = outer_c - outer(box_c, ignore_auto='I')
        min_c = box_c.minimum
        if inner_c >= min_c:
            # Ignore the preferred/maximum as the next best solution
            # is to drop that constraint.
            box_c.inner = inner_c
            return 'ok'
    # else: Over-constrained


@register(auto_inners=(0, 1, 0), auto_margins=(0, 0, 0))
def implementation_3(box_a, box_b, box_c, outer_sum):
    outer_a = outer(box_a)
    outer_c = outer(box_c)
    if outer_a == outer_c:  # Rule 2
        outer_b = outer_sum - outer_a - outer_c  # Rule 1
        inner_b = outer_b - outer(box_b, ignore_auto='I')
        min_b = box_b.minimum
        if inner_b >= min_b:
            # Ignore the preferred/maximum as the next best solution
            # is to drop that constraint.
            box_b.inner = inner_b
            return 'ok'
    # else: Over-constrained


@register(auto_inners=(0, 1, 1), auto_margins=(0, 0, 0))
def implementation_4(box_a, box_b, box_c, outer_sum):
    outer_a = outer(box_a)
    outer_c = outer_a  # Rule 2
    outer_b = outer_sum - outer_a - outer_c  # Rule 1

    inner_c = outer_c - outer(box_c, ignore_auto='I')
    inner_b = outer_b - outer(box_b, ignore_auto='I')

    min_c = box_c.minimum
    min_b = box_b.minimum
    if inner_b >= min_b and inner_c >= min_c:
        # Ignore the preferred/maximum as the next best solution
        # is to drop these constraints.
        box_b.inner = inner_b
        box_c.inner = inner_c
        return 'ok'
    # else: Over-constrained


@register(auto_inners=(1, 0, 1), auto_margins=(0, 0, 0))
def implementation_5(box_a, box_b, box_c, outer_sum):
    outer_b = outer(box_b)
    outer_ac = outer_sum - outer_b  # Rule 1
    outer_a = outer_ac / 2  # Rule 2
    outer_c = outer_ac / 2  # Rule 2

    inner_a = outer_a - outer(box_a, ignore_auto='I')
    inner_c = outer_c - outer(box_c, ignore_auto='I')

    min_a = box_a.minimum
    min_c = box_c.minimum
    if inner_a >= min_a and inner_c >= min_c:
        # Ignore the preferred/maximum as the next best solution
        # is to drop these constraints.
        box_a.inner = inner_a
        box_c.inner = inner_c
        return 'ok'
    # else: Over-constrained


@register(auto_inners=(1, 1, 1), auto_margins=(0, 0, 0))
def implementation_6(box_a, box_b, box_c, outer_sum):
    constants_a = outer(box_a, ignore_auto='I')
    constants_b = outer(box_b, ignore_auto='I')
    constants_c = outer(box_c, ignore_auto='I')

    min_outer_a = box_a.minimum + constants_a
    min_outer_b = box_b.minimum + constants_b
    min_outer_c = box_c.minimum + constants_c

    max_a = box_a.preferred
    max_b = box_b.preferred
    max_c = box_c.preferred
    max_outer_a = max_a + constants_a
    max_outer_b = max_b + constants_b
    max_outer_c = max_c + constants_c

    # Rule 2: Same outer for A and C, rule 2
    # So their common min/max is the most restrictive
    min_outer_ac = max(min_outer_a, min_outer_c)
    max_outer_ac = min(max_outer_a, max_outer_c)
    # Condition 1
    if outer_sum >= (2 * min_outer_ac + min_outer_b):
        weight_ac = max(max_a, max_c)
        weight_b = max_b
        total_weight = 2 * weight_ac + weight_b  # Rule 1
        if total_weight in (0, float('inf')):
            normalized_weight_b = 1 / 3
        else:
            normalized_weight_b = weight_b / total_weight
        inner_sum = outer_sum - constants_a - constants_b - constants_c
        inner_b = inner_sum * normalized_weight_b
        outer_b = inner_b + constants_b

        new_max_outer_b = outer_sum - 2 * min_outer_ac
        # min_outer_b <= max_outer_b is ensured by Condition 1
        if (
            outer_sum <= 2 * max_outer_ac + max_outer_b and
            max_outer_a >= min_outer_c and
            max_outer_c >= min_outer_a
        ):
            # outer_ac <= max_outer_ac
            # => (outer_sum - outer_b) / 2 <= max_outer_ac
            # => outer_b >= outer_sum - 2 * max_outer_ac
            min_outer_b = max(min_outer_b, outer_sum - 2 * max_outer_ac)
            max_outer_b = min(max_outer_b, new_max_outer_b)
        else:
            # Drop max constraints
            max_outer_b = new_max_outer_b
        assert min_outer_b <= max_outer_b
        outer_b = max(outer_b, min_outer_b)
        outer_b = min(outer_b, max_outer_b)
        # Rule 1 and 2
        outer_ac = (outer_sum - outer_b) / 2
        box_a.inner = outer_ac - constants_a
        box_b.inner = outer_b - constants_b
        box_c.inner = outer_ac - constants_c

        return 'ok'

    # Over-constrained by minimums.
    # Margins that become 'auto' will end up negative.
    # Choose inner so that margins are as small (close to 0) as possible.
    box_a.inner = min_outer_ac - constants_a
    box_b.inner = min_outer_b - constants_b
    box_c.inner = min_outer_ac - constants_c


@register(auto_inners=(0, 0, 0), auto_margins=(0, 0, 1))
def implementation_7(box_a, box_b, box_c, outer_sum):
    outer_a = outer(box_a)
    outer_b = outer(box_b)
    if outer_a == (outer_sum - outer_b) / 2: # Rules 1 and 2
        outer_c = outer_a  # Rule 2
        distribute_margins(box_c, outer_c)
        return 'ok'
    # else: Over-constrained


@register(auto_inners=(0, 0, 1), auto_margins=(0, 0, 1))
def implementation_8(box_a, box_b, box_c, outer_sum):
    outer_a = outer(box_a)
    outer_b = outer(box_b)
    if outer_a == (outer_sum - outer_b) / 2: # Rules 1 and 2
        outer_c = outer_a  # Rule 2
        distribute_margins_and_inner(box_c, outer_c)
        return 'ok'
    # else: Over-constrained


@register(auto_inners=(0, 1, 0), auto_margins=(0, 0, 1))
def implementation_9(box_a, box_b, box_c, outer_sum):
    outer_a = outer(box_a)
    outer_c = outer_a  # Rule 2
    outer_b = outer_sum - outer_a - outer_c  # Rule 1

    inner_b = outer_b - outer(box_b, ignore_auto='I')
    min_b = box_b.minimum
    if inner_b >= min_b:
        # Ignore the preferred/maximum as the next best solution
        # is to drop that constraints.
        box_b.inner = inner_b
        distribute_margins(box_c, outer_c)
        return 'ok'
    # else: Over-constrained


@register(auto_inners=(0, 1, 1), auto_margins=(0, 0, 1))
def implementation_10(box_a, box_b, box_c, outer_sum):
    outer_a = outer(box_a)
    outer_c = outer_a  # Rule 2
    outer_b = outer_sum - outer_a - outer_c  # Rule 1

    inner_b = outer_b - outer(box_b, ignore_auto='I')

    min_b = box_b.minimum
    if inner_b >= min_b:
        # Ignore the preferred/maximum as the next best solution
        # is to drop these constraints.
        box_b.inner = inner_b
        distribute_margins_and_inner(box_c, outer_c)
        return 'ok'
    # else: Over-constrained


@register(auto_inners=(1, 0, 0), auto_margins=(0, 0, 1))
def implementation_11(box_a, box_b, box_c, outer_sum):
    outer_b = outer(box_b)
    outer_ac = outer_sum - outer_b  # Rule 1
    outer_a = outer_ac / 2  # Rule 2
    outer_c = outer_ac / 2  # Rule 2

    inner_a = outer_a - outer(box_a, ignore_auto='I')
    min_a = box_a.minimum
    if inner_a >= min_a:
        # Ignore the preferred/maximum as the next best solution
        # is to drop these constraints.
        box_a.inner = inner_a
        distribute_margins(box_c, outer_c)
        return 'ok'
    # else: Over-constrained


@register(auto_inners=(1, 0, 1), auto_margins=(0, 0, 1))
def implementation_12(box_a, box_b, box_c, outer_sum):
    outer_b = outer(box_b)
    outer_ac = outer_sum - outer_b  # Rule 1
    outer_a = outer_ac / 2  # Rule 2
    outer_c = outer_ac / 2  # Rule 2

    inner_a = outer_a - outer(box_a, ignore_auto='I')
    min_a = box_a.minimum
    if inner_a >= min_a:
        # Ignore the preferred/maximum as the next best solution
        # is to drop these constraints.
        box_a.inner = inner_a

        remaining_c = outer_c - outer(box_c, ignore_auto='MI')
        set_inner_within_range(box_c, remaining_c)
        distribute_margins(box_c, outer_c)
        return 'ok'
    # else: Over-constrained


@register(auto_inners=(1, 1, 0), auto_margins=(0, 0, 1))
def implementation_13(box_a, box_b, box_c, outer_sum):
    constants_a = outer(box_a, ignore_auto='I')
    constants_b = outer(box_b, ignore_auto='I')
    constants_c = outer(box_c, ignore_auto='M')

    min_outer_a = box_a.minimum + constants_a
    min_outer_b = box_b.minimum + constants_b

    # C can have any margin, so it is unconstrained other than rule 2.
    if outer_sum >= min_outer_b + 2 * min_outer_a:
        # Ignore the preferred/maximum as the next best solution
        # is to drop these constraints.

        # Rule 2: Same outer for A and C, rule 2
        # So their common min/max is the most restrictive
        min_outer_ac = min_outer_a
        max_outer_ac = (outer_sum - min_outer_b) / 2

        # The goal function is minimized (auto margins == 0) for this value:
        optimal_outer_ac = constants_c

        # Within bounds
        outer_ac = min(max_outer_ac, max(min_outer_ac, optimal_outer_ac))
        outer_b = outer_sum - 2 * outer_ac

        box_a.inner = outer_ac - constants_a
        box_b.inner = outer_b - constants_b
        distribute_margins(box_c, outer_ac)
        return 'ok'
    # else: Over-constrained


@register(auto_inners=(1, 1, 1), auto_margins=(0, 0, 1))
def implementation_14(box_a, box_b, box_c, outer_sum):
    return NotImplemented


@register(auto_inners=(0, 0, 0), auto_margins=(0, 1, 0))
def implementation_15(box_a, box_b, box_c, outer_sum):
    outer_a = outer(box_a)
    outer_c = outer(box_c)
    if outer_a == outer_c:  # Rule 2
        outer_b = outer_sum - outer_a - outer_c  # Rule 1
        distribute_margins(box_b, outer_b)
        return 'ok'
    # else: Over-constrained


@register(auto_inners=(0, 0, 1), auto_margins=(0, 1, 0))
def implementation_16(box_a, box_b, box_c, outer_sum):
    outer_a = outer(box_a)
    outer_c = outer_a  # Rule 2
    outer_b = outer_sum - outer_a - outer_c  # Rule 1

    inner_c = outer_c - outer(box_c, ignore_auto='I')
    min_c = box_c.minimum
    if inner_c >= min_c:
        # Ignore the preferred/maximum as the next best solution
        # is to drop these constraints.
        box_c.inner = inner_c
        distribute_margins(box_b, outer_b)
        return 'ok'
    # else: Over-constrained


@register(auto_inners=(0, 1, 0), auto_margins=(0, 1, 0))
def implementation_17(box_a, box_b, box_c, outer_sum):
    outer_a = outer(box_a)
    outer_c = outer(box_c)
    if outer_a == outer_c:  # Rule 2
        outer_b = outer_sum - outer_a - outer_c  # Rule 1
        distribute_margins_and_inner(box_b, outer_b)
        return 'ok'
    # else: Over-constrained


@register(auto_inners=(0, 1, 1), auto_margins=(0, 1, 0))
def implementation_18(box_a, box_b, box_c, outer_sum):
    outer_a = outer(box_a)
    outer_c = outer_a  # Rule 2
    outer_b = outer_sum - outer_a - outer_c  # Rule 1

    inner_c = outer_c - outer(box_c, ignore_auto='I')
    min_c = box_c.minimum
    if inner_c >= min_c:
        # Ignore the preferred/maximum as the next best solution
        # is to drop these constraints.
        box_c.inner = inner_c
        distribute_margins_and_inner(box_b, outer_b)
        return 'ok'
    # else: Over-constrained


@register(auto_inners=(1, 0, 1), auto_margins=(0, 1, 0))
def implementation_19(box_a, box_b, box_c, outer_sum):
    constants_a = outer(box_a, ignore_auto='I')
    constants_b = outer(box_b, ignore_auto='M')
    constants_c = outer(box_c, ignore_auto='I')

    min_outer_a = box_a.minimum + constants_a
    min_outer_c = box_c.minimum + constants_c

    # Ignore the preferred/maximum as the next best solution
    # is to drop these constraints.

    # Rule 2: Same outer for A and C, rule 2
    # So their common min is the most restrictive
    min_outer_ac = max(min_outer_a, min_outer_c)

    # The goal function is minimized (auto margins == 0) for this value:
    optimal_outer_ac = (outer_sum - constants_b) / 2

    # Within bounds
    outer_ac = max(min_outer_ac, optimal_outer_ac)
    outer_b = outer_sum - 2 * outer_ac

    box_a.inner = outer_ac - constants_a
    box_c.inner = outer_ac - constants_c
    distribute_margins(box_b, outer_b)
    return 'ok'


@register(auto_inners=(1, 1, 1), auto_margins=(0, 1, 0))
def implementation_20(box_a, box_b, box_c, outer_sum):
    return NotImplemented


@register(auto_inners=(0, 0, 0), auto_margins=(0, 1, 1))
def implementation_21(box_a, box_b, box_c, outer_sum):
    outer_a = outer(box_a)
    outer_c = outer_a  # Rule 2
    outer_b = outer_sum - outer_a - outer_c  # Rule 1

    distribute_margins(box_b, outer_b)
    distribute_margins(box_c, outer_c)
    return 'ok'


@register(auto_inners=(0, 0, 1), auto_margins=(0, 1, 1))
def implementation_22(box_a, box_b, box_c, outer_sum):
    outer_a = outer(box_a)
    outer_c = outer_a  # Rule 2
    outer_b = outer_sum - outer_a - outer_c  # Rule 1

    distribute_margins(box_b, outer_b)
    distribute_margins_and_inner(box_c, outer_c)
    return 'ok'


@register(auto_inners=(0, 1, 0), auto_margins=(0, 1, 1))
def implementation_23(box_a, box_b, box_c, outer_sum):
    outer_a = outer(box_a)
    outer_c = outer_a  # Rule 2
    outer_b = outer_sum - outer_a - outer_c  # Rule 1

    distribute_margins_and_inner(box_b, outer_b)
    distribute_margins(box_c, outer_c)
    return 'ok'


@register(auto_inners=(0, 1, 1), auto_margins=(0, 1, 1))
def implementation_24(box_a, box_b, box_c, outer_sum):
    outer_a = outer(box_a)
    outer_c = outer_a  # Rule 2
    outer_b = outer_sum - outer_a - outer_c  # Rule 1

    distribute_margins_and_inner(box_b, outer_b)
    distribute_margins_and_inner(box_c, outer_c)
    return 'ok'


@register(auto_inners=(1, 0, 0), auto_margins=(0, 1, 1))
def implementation_25(box_a, box_b, box_c, outer_sum):
    constants_a = outer(box_a, ignore_auto='I')
    constants_b = outer(box_b, ignore_auto='M')
    constants_c = outer(box_c, ignore_auto='M')

    n_margins_b = [box_b.margin_a, box_b.margin_b].count('auto')
    n_margins_c = [box_c.margin_a, box_c.margin_b].count('auto')

    # Math:
    # Use rule 1 and 2 to rewrite the goal function in terms of outer_ac
    # The result is still quadratic but with only one variable:
    #    Goal = A * outer_ac**2  +  B * outer_ac  +  C
    # The minimum of a quadratic function is at
    #    outer_ac = -B / (2 * A)
    optimal_outer_ac = (
          constants_c / n_margins_c
        + 2 * (outer_sum - constants_b) / n_margins_b
    ) / (
          1 / n_margins_c
        + 4 / n_margins_b
    )
    optimal_inner_a = optimal_outer_ac - constants_a

    set_inner_within_range(box_a, optimal_inner_a)

    outer_ac = box_a.inner + constants_a
    outer_b = outer_sum - 2 * outer_ac
    distribute_margins(box_b, outer_b)
    distribute_margins(box_c, outer_ac)
    return 'ok'


@register(auto_inners=(1, 0, 1), auto_margins=(0, 1, 1))
def implementation_26(box_a, box_b, box_c, outer_sum):
    return NotImplemented


@register(auto_inners=(1, 1, 0), auto_margins=(0, 1, 1))
def implementation_27(box_a, box_b, box_c, outer_sum):
    return NotImplemented


@register(auto_inners=(1, 1, 1), auto_margins=(0, 1, 1))
def implementation_28(box_a, box_b, box_c, outer_sum):
    return NotImplemented


@register(auto_inners=(0, 0, 0), auto_margins=(1, 0, 1))
def implementation_29(box_a, box_b, box_c, outer_sum):
    new_outer_ac = (outer_sum - outer(box_b)) / 2
    distribute_margins(box_a, new_outer_ac)
    distribute_margins(box_c, new_outer_ac)
    return 'ok'


@register(auto_inners=(0, 0, 1), auto_margins=(1, 0, 1))
def implementation_30(box_a, box_b, box_c, outer_sum):
    outer_b = outer(box_b)
    outer_ac = outer_sum - outer_b  # Rule 1
    outer_a = outer_ac / 2  # Rule 2
    outer_c = outer_ac / 2  # Rule 2

    distribute_margins(box_a, outer_a)
    distribute_margins_and_inner(box_c, outer_c)
    return 'ok'


@register(auto_inners=(0, 1, 0), auto_margins=(1, 0, 1))
def implementation_31(box_a, box_b, box_c, outer_sum):
    constants_a = outer(box_a, ignore_auto='M')
    constants_b = outer(box_b, ignore_auto='I')
    constants_c = outer(box_c, ignore_auto='M')

    n_margins_a = [box_a.margin_a, box_a.margin_b].count('auto')
    n_margins_c = [box_c.margin_a, box_c.margin_b].count('auto')

    # Math:
    # Use rule 1 and 2 to rewrite the goal function in terms of outer_ac
    # The result is still quadratic but with only one variable:
    #    Goal = A * outer_ac**2  +  B * outer_ac  +  C
    # The minimum of a quadratic function is at
    #    outer_ac = -B / (2 * A)
    optimal_outer_ac = (
          constants_a / n_margins_a
        + constants_c / n_margins_c
    ) / (
          1 / n_margins_a
        + 1 / n_margins_c
    )
    optimal_outer_b = outer_sum - 2 * optimal_outer_ac
    optimal_inner_b = optimal_outer_b - constants_b

    set_inner_within_range(box_b, optimal_inner_b)

    outer_b = box_b.inner + constants_b
    outer_ac = (outer_sum - outer_b) / 2
    distribute_margins(box_a, outer_ac)
    distribute_margins(box_c, outer_ac)
    return 'ok'


@register(auto_inners=(0, 1, 1), auto_margins=(1, 0, 1))
def implementation_32(box_a, box_b, box_c, outer_sum):
    return NotImplemented


@register(auto_inners=(1, 0, 1), auto_margins=(1, 0, 1))
def implementation_33(box_a, box_b, box_c, outer_sum):
    outer_b = outer(box_b)
    outer_ac = outer_sum - outer_b  # Rule 1
    outer_a = outer_ac / 2  # Rule 2
    outer_c = outer_ac / 2  # Rule 2

    distribute_margins_and_inner(box_a, outer_a)
    distribute_margins_and_inner(box_c, outer_c)
    return 'ok'


@register(auto_inners=(1, 1, 1), auto_margins=(1, 0, 1))
def implementation_34(box_a, box_b, box_c, outer_sum):
    return NotImplemented


@register(auto_inners=(0, 0, 0), auto_margins=(1, 1, 1))
def implementation_35(box_a, box_b, box_c, outer_sum):
    constants_a = outer(box_a, ignore_auto='M')
    constants_b = outer(box_b, ignore_auto='M')
    constants_c = outer(box_c, ignore_auto='M')

    n_margins_a = [box_a.margin_a, box_a.margin_b].count('auto')
    n_margins_b = [box_b.margin_a, box_b.margin_b].count('auto')
    n_margins_c = [box_c.margin_a, box_c.margin_b].count('auto')

    # Math:
    # Use rule 1 and 2 to rewrite the goal function in terms of outer_ac
    # The result is still quadratic but with only one variable:
    #    Goal = A * outer_ac**2  +  B * outer_ac  +  C
    # The minimum of a quadratic function is at
    #    outer_ac = -B / (2 * A)
    outer_ac = (
          constants_a / n_margins_a
        + constants_c / n_margins_c
        + 2 * (outer_sum - constants_b) / n_margins_b
    ) / (
          1 / n_margins_a
        + 1 / n_margins_b
        + 1 / n_margins_c
    )
    outer_b = outer_sum - 2 * outer_ac
    distribute_margins(box_a, outer_ac)
    distribute_margins(box_b, outer_b)
    distribute_margins(box_c, outer_ac)
    return 'ok'



@register(auto_inners=(0, 0, 1), auto_margins=(1, 1, 1))
def implementation_36(box_a, box_b, box_c, outer_sum):
    return NotImplemented


@register(auto_inners=(0, 1, 0), auto_margins=(1, 1, 1))
def implementation_37(box_a, box_b, box_c, outer_sum):
    return NotImplemented


@register(auto_inners=(0, 1, 1), auto_margins=(1, 1, 1))
def implementation_38(box_a, box_b, box_c, outer_sum):
    return NotImplemented


@register(auto_inners=(1, 0, 1), auto_margins=(1, 1, 1))
def implementation_39(box_a, box_b, box_c, outer_sum):
    return NotImplemented


@register(auto_inners=(1, 1, 1), auto_margins=(1, 1, 1))
def implementation_40(box_a, box_b, box_c, outer_sum):
    return NotImplemented

assert len(IMPLEMENTATIONS) == 64



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Module managing CSS.

This module takes care of steps 3 and 4 of “CSS 2.1 processing model”:
Retrieve stylesheets associated with a document and annotate every element
with a value for every CSS property.
http://www.w3.org/TR/CSS21/intro.html#processing-model

This module does this in more than two steps. The
:func:`get_all_computed_styles` function does everything, but there is also a
function for each step:

- ``find_stylesheets``: Find and parse all author stylesheets in a document
- ``effective_rules``: Resolve @media and @import rules
- ``match_selectors``: Find elements in a document that match a selector list
- ``find_style_attributes``: Find and parse all `style` HTML attributes
- ``effective_declarations``: Remove ignored properties and expand shorthands
- ``add_property``: Take applicable properties and only keep those with
  highest weight.
- ``set_computed_styles``: Handle initial values, inheritance and computed
  values for one element.

"""

from __future__ import division, unicode_literals

import re
import logging
import os.path

from lxml import cssselect
import cssutils

from .colors import CSS3_COLORS
from . import properties
from . import validation
from . import computed_values
from ..utils import get_url_attribute, urllib_fetcher
from ..logger import LOGGER
from ..compat import iteritems


# Monkey-patch cssutils to add extended color keywords:
cssutils.css.ColorValue.COLORS.update(CSS3_COLORS)


PARSER = cssutils.CSSParser(parseComments=False, validate=False)
PARSER.setFetcher(urllib_fetcher)

HTML5_UA_STYLESHEET = PARSER.parseFile(
    os.path.join(os.path.dirname(__file__), 'html5_ua.css'))


# Pseudo-classes and pseudo-elements are the same to lxml.cssselect.parse().
# List the identifiers for all CSS3 pseudo elements here to distinguish them.
PSEUDO_ELEMENTS = ('before', 'after', 'first-line', 'first-letter')

# Selectors for @page rules can have a pseudo-class, one of :first, :left
# or :right. This maps pseudo-classes to lists of "page types" selected.
PAGE_PSEUDOCLASS_TARGETS = {
    ':first': ['first_left_page', 'first_right_page'],
    ':left': ['left_page', 'first_left_page'],
    ':right': ['right_page', 'first_right_page'],
    # no pseudo-class: all pages
    None: ['left_page', 'right_page', 'first_left_page', 'first_right_page'],
}

# A test function that returns True if the given property name has an
# initial value that is not always the same when computed.
RE_INITIAL_NOT_COMPUTED = re.compile(
    '^(display|border_[a-z]+_(width|color))$').match


class StyleDict(object):
    """A mapping (dict-like) that allows attribute access to values.

    Allow eg. ``style.font_size`` instead of ``style['font-size']``.

    :param parent: if given, should be a mapping. Values missing from this
                   dict will be looked up in the parent dict. Setting a value
                   in this dict masks any value in the parent.

    """
    def __init__(self, data=None, parent=None):
        if data is None:
            data = {}
        else:
            data = dict(data)
        if parent is None:
            parent = {}
        # work around our own __setattr__
        object.__setattr__(self, '_storage', data)
        object.__setattr__(self, '_parent', parent)

    def __getitem__(self, key):
        storage = self._storage
        if key in storage:
            return storage[key]
        else:
            return self._parent[key]

    def __setitem__(self, key, value):
        self._storage[key] = value

    def updated_copy(self, other):
        copy = self.copy()
        copy._storage.update(other)
        return copy

    def __contains__(self, key):
        return key in self._parent or key in self._storage

    __getattr__ = __getitem__  # May raise KeyError instead of AttributeError
    __setattr__ = __setitem__

    def copy(self):
        """Copy the ``StyleDict``.

        Create a new StyleDict with this one as the parent. This is a cheap
        "copy-on-write". Modifications in the copy will not affect
        the original, but modifications in the original *may* affect the
        copy.

        """
        if self._storage:
            parent = self
        else:
            parent = self._parent
        style = type(self)(parent=parent)
        if self.anonymous:
            object.__setattr__(style, 'anonymous', True)
        return style

    def as_dict(self):
        """Return a new dict completly separate from this object."""
        if hasattr(self._parent, 'as_dict'):
            rv = self._parent.as_dict()
        elif hasattr(self._parent, 'items'):
            rv = dict(self._parent)
        else:
            rv = {}
        rv.update(self._storage)
        return rv

    def inherit_from(self):
        """Return a new StyleDict with inherited properties from this one.

        Non-inherited properties get their initial values.
        This is the styles for an anonymous box.
        """
        style = computed_from_cascaded(cascaded={}, parent_style=self,
            # Only used by non-inherited properties. eg `content: attr(href)`
            element=None)
        object.__setattr__(style, 'anonymous', True)
        return style

    # Default values, may be overriden on instances
    anonymous = False


def find_stylesheets(document):
    """Yield the stylesheets of ``document``.

    The output order is the same as the order of the dom.

    """
    parser = PARSER
    for element in document.dom.iter():
        if element.tag not in ('style', 'link'):
            continue
        mimetype = element.get('type')
        # Only keep 'type/subtype' from 'type/subtype ; param1; param2'.
        if mimetype and mimetype.split(';', 1)[0].strip() != 'text/css':
            continue
        # cssutils translates '' to 'all'.
        media_attr = (element.get('media') or '').strip()
        if element.tag == 'style':
            # TODO: handle the `scoped` attribute
            # Content is text that is directly in the <style> element, not its
            # descendants
            content = [element.text or '']
            for child in element:
                content.append(child.tail or '')
            content = ''.join(content)
            # lxml should give us either unicode or ASCII-only bytestrings, so
            # we don't need `encoding` here.
            yield parser.parseString(content,
                href=element.base_url,
                media=media_attr, title=element.get('title'))
        elif element.tag == 'link' and element.get('href'):
            rel = element.get('rel', '').split()
            if 'stylesheet' not in rel or 'alternate' in rel:
                continue
            # URLs should NOT have been made absolute earlier
            # TODO: support the <base> HTML element, but do not use
            # lxml.html.HtmlElement.make_links_absolute() that changes the tree
            href = get_url_attribute(element, 'href')
            yield parser.parseUrl(
                href, media=media_attr, title=element.get('title'))


def find_style_attributes(document):
    """Yield the ``element, declaration_block`` of ``document``."""
    parser = PARSER
    for element in document.dom.iter():
        style_attribute = element.get('style')
        if style_attribute:
            # TODO: no href for parseStyle. What about relative URLs?
            # CSS3 says we should resolve relative to the attribute:
            # http://www.w3.org/TR/css-style-attr/#interpret
            yield element, parser.parseStyle(style_attribute)


def evaluate_media_query(query_list, medium):
    """Return the boolean evaluation of `query_list` for the given `medium`.

    :attr query_list: a cssutilts.stlysheets.MediaList
    :attr medium: a media type string (for now)

    """
    # TODO: actual support for media queries, not just media types
    return query_list.mediaText == 'all' \
        or any(query.mediaText == medium for query in query_list)


def effective_rules(sheet, medium):
    """Yield applicable rules of ``sheet`` for ``medium``.

    The rules include those defined with ``@import`` and ``@media`` rules.

    """
    # sheet.media is not intrinsic but comes from where the stylesheet was
    # found: media HTML attribute, @import or @media rule.
    if not evaluate_media_query(sheet.media, medium):
        return
    for rule in sheet.cssRules:
        if rule.type in (rule.CHARSET_RULE, rule.COMMENT):
            continue  # ignored
        elif rule.type == rule.IMPORT_RULE:
            subsheet = rule.styleSheet
        elif rule.type == rule.MEDIA_RULE:
            # CSSMediaRule is kinda like a CSSStyleSheet: it has media and
            # cssRules attributes.
            subsheet = rule
        else:
            # pass other rules through: "normal" rulesets, @font-face,
            # @namespace, @page, and @variables
            yield rule
            if rule.type == rule.PAGE_RULE:
                # Do not use recursion here as page rules have no .media attr
                for margin_rule in rule.cssRules:
                    yield margin_rule
            continue  # no sub-stylesheet here.
        # subsheet has the .media attribute from the @import or @media rule.
        for subrule in effective_rules(subsheet, medium):
            yield subrule


def effective_declarations(declaration_block):
    """Yield ``property_name, property_value_list, importance`` tuples.

    In the given ``declaration_block``, the invalid or unsupported declarations
    are ignored and the shorthand properties are expanded.

    """
    for declaration in declaration_block.getProperties(all=True):
        for name, values in validation.validate_and_expand(
                declaration.name.replace('-', '_'),
                # list() may call len(), which is slow on PropertyValue
                # Use iter() to avoid this.
                list(iter(declaration.propertyValue))):
            yield name, (values, declaration.priority)


def declaration_precedence(origin, importance):
    """Return the precedence for a declaration.

    Precedence values have no meaning unless compared to each other.

    Acceptable values for ``origin`` are the strings ``'author'``, ``'user'``
    and ``'user agent'``.

    """
    # See http://www.w3.org/TR/CSS21/cascade.html#cascading-order
    if origin == 'user agent':
        return 1
    elif origin == 'user' and not importance:
        return 2
    elif origin == 'author' and not importance:
        return 3
    elif origin == 'author':  # and importance
        return 4
    elif origin == 'user':  # and importance
        return 5
    else:
        assert ValueError('Unkown origin: %r' % origin)


def add_declaration(cascaded_styles, prop_name, prop_values, weight, element,
                    pseudo_type=None):
    """Set the value for a property on a given element.

    The value is only set if there is no value of greater weight defined yet.

    """
    style = cascaded_styles.setdefault((element, pseudo_type), {})
    _values, previous_weight = style.get(prop_name, (None, None))
    if previous_weight is None or previous_weight <= weight:
        style[prop_name] = prop_values, weight


def selector_to_xpath(selector):
    """Return ``pseudo_type, selector_callable`` from a cssutils ``selector``.

    ``pseudo_type`` is a string and ``selector_callable`` is a
    :class:`lxml.cssselect` XPath callable.

    """
    try:
        return selector._x_weasyprint_parsed_cssselect
    except AttributeError:
        parsed_selector = cssselect.parse(selector.selectorText)
        # cssutils made sure that `selector` is not a "group of selectors"
        # in CSS3 terms (`rule.selectorList` is) so `parsed_selector` cannot be
        # of type `cssselect.Or`.
        # This leaves only three cases:
        # - The selector ends with a pseudo-element. As `cssselect.parse()`
        #   parses left-to-right, `parsed_selector` is a `cssselect.Pseudo`
        #   instance that we can unwrap. This is the only place where CSS
        #   allows pseudo-element selectors.
        # - The selector has a pseudo-element not at the end. This is invalid
        #   and the whole ruleset should be ignored.
        #   cssselect.CSSSelector() will raise a cssselect.ExpressionError.
        # - The selector has no pseudo-element and is supported by
        #   `cssselect.CSSSelector`.
        if isinstance(parsed_selector, cssselect.CombinedSelector):
            simple_selector = parsed_selector.subselector
            if isinstance(simple_selector, cssselect.Pseudo) \
                    and simple_selector.ident in PSEUDO_ELEMENTS:
                pseudo_type = str(simple_selector.ident)
                # Remove the pseudo-element from the selector
                parsed_selector.subselector = simple_selector.element
            else:
                # No pseudo-element or invalid selector.
                pseudo_type = None
        else:
            if isinstance(parsed_selector, cssselect.Pseudo) \
                    and parsed_selector.ident in PSEUDO_ELEMENTS:
                pseudo_type = str(parsed_selector.ident)
                # Remove the pseudo-element from the selector
                parsed_selector = parsed_selector.element
            else:
                # No pseudo-element or invalid selector.
                pseudo_type = None

        selector_callable = cssselect.CSSSelector(parsed_selector)
        result = (pseudo_type, selector_callable)

        # Cache for next time we use the same stylesheet
        selector._x_weasyprint_parsed_cssselect = result
        return result


def match_selectors(document, rule):
    """Get the elements in ``document`` matched by the select selectors in
    ``rule``.

    Yield ``element, pseudo_element_type, selector_specificity`` tuples.

    ``selector_list`` should be an iterable of ``cssutils.Selector`` objects.

    If any of the selectors is invalid, an empty iterable is returned as the
    whole rule should be ignored.

    """
    selectors = []
    for selector in rule.selectorList:
        try:
            pseudo_type, selector_callable = selector_to_xpath(selector)
        except (cssselect.SelectorSyntaxError, cssselect.ExpressionError,
                NotImplementedError):
            LOGGER.warn('Unsupported selector `%s`, the whole rule-set '
                        'was ignored.', selector.selectorText)
            return
        selectors.append(
            (selector_callable, pseudo_type, selector.specificity))

    # Only apply to elements after seeing all selectors, as we want to
    # ignore he whole ruleset if just one selector is invalid.
    # TODO: test that ignoring actually happens.
    for selector, pseudo_type, specificity in selectors:
        for element in selector(document.dom):
            yield element, pseudo_type, specificity


def match_page_selector(rule, margin_type=None):
    """Get the page types matching the selector in ``rule``.

    Yield ``page_type, margin_type, selector_specificity`` tuples.
    ``margin_type`` is as passed as an argument.

    Return an empty iterable if the selector is invalid or unsupported.

    """
    selector = rule.selectorText
    # TODO: support "page names" in page selectors (see CSS3 Paged Media)
    pseudo_class = selector or None
    page_types = PAGE_PSEUDOCLASS_TARGETS.get(pseudo_class, None)
    if page_types is None:
        LOGGER.warn('Unsupported @page selector %r, the whole rule-set '
                    'was ignored.', selector)
    else:
        specificity = rule.specificity
        for page_type in page_types:
            yield page_type, margin_type, specificity


def set_computed_styles(cascaded_styles, computed_styles,
                        element, parent, pseudo_type=None):
    """Set the computed values of styles to ``element``.

    Take the properties left by ``apply_style_rule`` on an element or
    pseudo-element and assign computed values with respect to the cascade,
    declaration priority (ie. ``!important``) and selector specificity.

    """
    if parent is None:
        parent_style = None
    else:
        parent_style = computed_styles[parent, None]

    cascaded = cascaded_styles.get((element, pseudo_type), {})
    style = computed_from_cascaded(
        element, cascaded, parent_style, pseudo_type)
    computed_styles[element, pseudo_type] = style


def computed_from_cascaded(element, cascaded, parent_style, pseudo_type=None):
    """Get a dict of computed style mixed from parent and cascaded styles."""
    if not cascaded and parent_style is not None:
        # Fast path for anonymous boxes:
        # no cascaded style, only implicitly initial or inherited values.
        computed = StyleDict(parent=properties.INITIAL_VALUES)
        for name in properties.INHERITED:
            computed[name] = parent_style[name]
        # border-*-style is none, so border-width computes to zero.
        # Other than that, properties that would need computing are
        # border-*-color, but they do not apply.
        for side in ('top', 'bottom', 'left', 'right'):
            computed['border_%s_width' % side] = 0
        return computed

    # Handle inheritance and initial values
    specified = StyleDict()
    computed = StyleDict()
    for name, initial in iteritems(properties.INITIAL_VALUES):
        if name in cascaded:
            value, _precedence = cascaded[name]
            keyword = value
        else:
            if name in properties.INHERITED:
                keyword = 'inherit'
            else:
                keyword = 'initial'

        if keyword == 'inherit' and parent_style is None:
            # On the root element, 'inherit' from initial values
            keyword = 'initial'

        if keyword == 'initial':
            value = initial
            if not RE_INITIAL_NOT_COMPUTED(name):
                # The value is the same as when computed
                computed[name] = value
        elif keyword == 'inherit':
            value = parent_style[name]
            # Values in parent_style are already computed.
            computed[name] = value

        assert value is not None
        specified[name] = value

    return computed_values.compute(
        element, pseudo_type, specified, computed, parent_style)


def get_all_computed_styles(document, medium,
                            user_stylesheets=None, ua_stylesheets=None):
    """Compute all the computed styles of ``document`` for ``medium``.

    Do everything from finding author stylesheets in the given HTML document
    to parsing and applying them.

    Return a dict of (DOM element, pseudo element type) -> StyleDict instance.

    """
    author_stylesheets = [sheet for sheet in find_stylesheets(document)
                          # It seems that cssutils returns None for
                          # some invalid (non-css?) stylesheets.
                          if sheet is not None]

    # keys: (element, pseudo_element_type)
    #    element: a lxml element object or the '@page' string for @page styles
    #    pseudo_element_type: a string such as 'first' (for @page) or 'after',
    #        or None for normal elements
    # values: dicts of
    #     keys: property name as a string
    #     values: (values, weight)
    #         values: a PropertyValue-like object
    #         weight: values with a greater weight take precedence, see
    #             http://www.w3.org/TR/CSS21/cascade.html#cascading-order
    cascaded_styles = {}

    for sheets, origin in (
        # Order here is not important ('origin' is).
        # Use this order for a regression test
        (ua_stylesheets or [], 'user agent'),
        (author_stylesheets, 'author'),
        (user_stylesheets or [], 'user'),
    ):
        if origin == 'user agent':
            # XXX temporarily disable logging for user-agent stylesheet
            level = LOGGER.level
            LOGGER.setLevel(logging.ERROR)

        for sheet in sheets:
            # TODO: UA and maybe user stylesheets might only need to be
            # expanded once, not for every document.
            for rule in effective_rules(sheet, medium):
                if rule.type == rule.STYLE_RULE:
                    matched = match_selectors(document, rule)
                elif rule.type == rule.PAGE_RULE:
                    matched = match_page_selector(rule)
                elif rule.type == rule.MARGIN_RULE:
                    # TODO: refactor this to reuse the result of
                    # match_page_selector() on the parent @page rule.
                    matched = match_page_selector(rule.parent, rule.margin)
                else:
                    # TODO: handle @font-face, @namespace, and @variables
                    continue

                declarations = list(effective_declarations(rule.style))

                if not declarations:
                    # Don’t bother working for nuthin’
                    # ``matched`` is a generator, so no work is done until
                    # we start iterating it.
                    continue

                for element, pseudo_type, specificity in matched:
                    for name, (values, importance) in declarations:
                        precedence = declaration_precedence(origin, importance)
                        weight = (precedence, specificity)
                        add_declaration(cascaded_styles, name, values, weight,
                                        element, pseudo_type)

        if origin == 'user agent':
            LOGGER.setLevel(level)

    for element, declarations in find_style_attributes(document):
        for name, (values, importance) in effective_declarations(declarations):
            precedence = declaration_precedence('author', importance)
            # 1 for being a style attribute, 0 as there is no selector.
            weight = (precedence, (1, 0, 0, 0))
            add_declaration(cascaded_styles, name, values, weight, element)

    # keys: (element, pseudo_element_type), like cascaded_styles
    # values: StyleDict objects:
    #     keys: property name as a string
    #     values: a PropertyValue-like object
    computed_styles = {}

    # First, computed styles for "real" elements *in tree order*
    # Tree order is important so that parents have computed styles before
    # their children, for inheritance.

    # Iterate on all elements, even if there is no cascaded style for them.
    for element in document.dom.iter():
        set_computed_styles(cascaded_styles, computed_styles, element,
                            parent=element.getparent())


    # Then computed styles for @page.

    # Iterate on all possible page types, even if there is no cascaded style
    # for them.
    for page_type in PAGE_PSEUDOCLASS_TARGETS[None]:
        set_computed_styles(cascaded_styles, computed_styles, page_type,
        # @page inherits from the root element:
        # http://lists.w3.org/Archives/Public/www-style/2012Jan/1164.html
                            parent=document.dom)

    # Then computed styles for pseudo elements, in any order.
    # Pseudo-elements inherit from their associated element so they come
    # last. Do them in a second pass as there is no easy way to iterate
    # on the pseudo-elements for a given element with the current structure
    # of cascaded_styles. (Keys are (element, pseudo_type) tuples.)

    # Only iterate on pseudo-elements that have cascaded styles. (Others
    # might as well not exist.)
    for element, pseudo_type in cascaded_styles:
        if pseudo_type:
            set_computed_styles(cascaded_styles, computed_styles,
                                element, pseudo_type=pseudo_type,
                                # The pseudo-element inherits from the element.
                                parent=element)

    return computed_styles



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Python 2/3 compatibility.

"""

from __future__ import division, unicode_literals

import io
import sys
import email
import contextlib


if sys.version_info[0] >= 3:
    # Python 3
    from urllib.parse import urljoin, urlparse, unquote_to_bytes
    from urllib.request import urlopen, Request
    from array import array

    basestring = str
    xrange = range
    iteritems = dict.items


    def urlopen_contenttype(url):
        """Return (file_obj, mime_type, encoding)"""
        result = urlopen(url)
        info = result.info()
        mime_type = info.get_content_type()
        charset = info.get_param('charset')
        # Using here result.fp gives 'ValueError: read of closed file'
        return result, mime_type, charset


    def parse_email(data):
        if isinstance(data, bytes):
            data = data.decode('utf8')
        return email.message_from_string(data)

else:
    # Python 2
    from urlparse import urljoin, urlparse, unquote
    from urllib2 import urlopen, Request
    from array import array as _array

    basestring = basestring
    xrange = xrange
    iteritems = dict.iteritems


    def array(typecode, initializer):
        return _array(typecode.encode('ascii'), initializer)


    def urlopen_contenttype(url):
        """Return (file_obj, mime_type, encoding)"""
        result = urlopen(url)
        info = result.info()
        mime_type = info.gettype()
        charset = info.getparam('charset')
        return result.fp, mime_type, charset


    def unquote_to_bytes(data):
        if isinstance(data, unicode):
            data = data.encode('ascii')
        return unquote(data)


    def parse_email(data):
        if isinstance(data, unicode):
            data = data.encode('utf8')
        return email.message_from_string(data)



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Functions laying out the markers.

"""

from __future__ import division, unicode_literals

import cairo

from .percentages import resolve_percentages
from ..text import TextFragment
from ..formatting_structure import boxes


def list_marker_layout(document, box, containing_block):
    """Lay out the list markers of ``box``."""
    # List markers can be either 'inside' or 'outside'.
    # Inside markers are layed out just like normal inline content, but
    # outside markers need specific layout.
    # TODO: implement outside markers in terms of absolute positioning,
    # see CSS3 lists.
    marker = getattr(box, 'outside_list_marker', None)
    if marker:
        resolve_percentages(marker, containing_block)
        if isinstance(marker, boxes.TextBox):
            text_fragment = TextFragment(marker.text, marker.style,
                context=cairo.Context(document.surface))
            result = text_fragment.split_first_line()
            (marker.show_line, _, marker.width, marker.height,
                marker.baseline, _) = result
        else:
            # Image marker
            image_marker_layout(marker)

        # Align the top of the marker box with the top of its list-item’s
        # content-box.
        # TODO: align the baselines of the first lines instead?
        marker.position_y = box.content_box_y()
        # ... and its right with the left of its list-item’s padding box.
        # (Swap left and right for right-to-left text.)
        marker.position_x = box.border_box_x()

        half_em = 0.5 * box.style.font_size
        direction = box.style.direction
        if direction == 'ltr':
            marker.margin_right = half_em
            marker.position_x -= marker.margin_width()
        else:
            marker.margin_left = half_em
            marker.position_x += box.border_width()


def image_marker_layout(box):
    """Layout the :class:`boxes.ImageMarkerBox` ``box``.

    :class:`boxes.ImageMarkerBox` objects are :class:`boxes.ReplacedBox`
    objects, but their used size is computed differently.

    """
    image, width, height = box.replacement
    ratio = width / height
    one_em = box.style.font_size
    if width is not None and height is not None:
        box.width = width
        box.height = height
    elif width is not None and ratio is not None:
        box.width = width
        box.height = width / ratio
    elif height is not None and ratio is not None:
        box.width = height * ratio
        box.height = height
    elif ratio is not None:
        # ratio >= 1 : width >= height
        if ratio >= 1:
            box.width = one_em
            box.height = one_em / ratio
        else:
            box.width = one_em * ratio
            box.height = one_em
    else:
        box.width = width if width is not None else one_em
        box.height = height if height is not None else one_em



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
WeasyPrint testing suite.

"""


from __future__ import division, unicode_literals



# coding: utf8

#  WeasyPrint converts web documents (HTML, CSS, ...) to PDF.
#  Copyright (C) 2011  Simon Sapin
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Affero General Public License as
#  published by the Free Software Foundation, either version 3 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Affero General Public License for more details.
#
#  You should have received a copy of the GNU Affero General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Implement the various counter types and list-style-type values.

These are defined in the same terms as CSS 3 Lists:
http://dev.w3.org/csswg/css3-lists/#predefined-counters

"""

from __future__ import division, unicode_literals

import functools

__all__ = ['format', 'format_list_marker']




# Initial values for counter style descriptors.
INITIAL_VALUES = dict(
    negative=('-', ''),
    prefix='',
    suffix='.',
    range=(float('-inf'), float('inf')),
    fallback='decimal',
    # type and symbols ommited here.
)

# Maps counter-style names to a dict of descriptors.
STYLES = {
    # Included here for format_list_marker().
    # format() special-cases decimal and does not use this.
    'decimal': INITIAL_VALUES,
}

# Maps counter types to a function implementing it.
# The functions take three arguments: the values of the `symbols`
# (or `additive-symbols` for the additive type) and `negative` descriptors,
# and the integer value being formatted.
# They return the representation as a string or None. None means that
# the value can not represented and the fallback should be used.
FORMATTERS = {}


def register_style(name, type='symbolic', **descriptors):
    """Register a counter style."""
    if type == 'override':
        # TODO: when @counter-style rules are supported, change override
        # to bind when a value is generated, not when the @rule is parsed.
        style = dict(STYLES[descriptors.pop('override')])
    else:
        style = dict(INITIAL_VALUES, formatter=functools.partial(
            FORMATTERS[type],
            descriptors.pop('symbols'),
            descriptors.pop('negative', INITIAL_VALUES['negative'])))
    style.update(descriptors)
    STYLES[name] = style


def register_formatter(function):
    """Register a counter type/algorithm."""
    FORMATTERS[function.__name__.replace('_', '-')] = function
    return function


@register_formatter
def repeating(symbols, _negative, value):
    """Implement the algorithm for `type: repeating`."""
    return symbols[(value - 1) % len(symbols)]


@register_formatter
def numeric(symbols, negative, value):
    """Implement the algorithm for `type: numeric`."""
    if value == 0:
        return symbols[0]
    is_negative = value < 0
    if is_negative:
        value = abs(value)
        prefix, suffix = negative
        reversed_parts = [suffix]
    else:
        reversed_parts = []
    length = len(symbols)
    value = abs(value)
    while value != 0:
        reversed_parts.append(symbols[value % length])
        value //= length
    if is_negative:
        reversed_parts.append(prefix)
    return ''.join(reversed(reversed_parts))


@register_formatter
def alphabetic(symbols, _negative, value):
    """Implement the algorithm for `type: alphabetic`."""
    if value <= 0:
        return None
    length = len(symbols)
    reversed_parts = []
    while value != 0:
        value -= 1
        reversed_parts.append(symbols[value % length])
        value //= length
    return ''.join(reversed(reversed_parts))


@register_formatter
def symbolic(symbols, _negative, value):
    """Implement the algorithm for `type: symbolic`."""
    if value <= 0:
        return None
    length = len(symbols)
    return symbols[value % length] * ((value - 1) // length)


@register_formatter
def non_repeating(symbols, _negative, value):
    """Implement the algorithm for `type: non-repeating`."""
    first_symbol_value, symbols = symbols
    value -= first_symbol_value
    if 0 <= value < len(symbols):
        return symbols[value]


@register_formatter
def additive(symbols, negative, value):
    """Implement the algorithm for `type: additive`."""
    if value == 0:
        for weight, symbol in symbols:
            if weight == 0:
                return symbol
    is_negative = value < 0
    if is_negative:
        value = abs(value)
        prefix, suffix = negative
        parts = [prefix]
    else:
        parts = []
    for weight, symbol in symbols:
        repetitions = value // weight
        parts.extend([symbol] * repetitions)
        value -= weight * repetitions
        if value == 0:
            if is_negative:
                parts.append(suffix)
            return ''.join(parts)
    return None  # Failed to find a representation for this value


# 'decimal' behaves the same as this, but defining it this way is silly.
# We’ll special-case it and just use str().
#register_style(
#    'decimal',
#    type='numeric',
#    symbols='0 1 2 3 4 5 6 7 8 9'.split(),
#)
register_style(
    'decimal-leading-zero',
    type='non-repeating',
    symbols=(-9, '''-09 -08 -07 -06 -05 -04 -03 -02 -01
                    00 01 02 03 04 05 06 07 08 09'''.split()),
)
register_style(
    'lower-roman',
    type='additive',
    range=(1, 4999),
    symbols=[(1000, 'm'), (900, 'cm'), (500, 'd'), (400, 'cd'),
             (100, 'c'), (90, 'xc'), (50, 'l'), (40, 'xl'),
             (10, 'x'), (9, 'ix'), (5, 'v'), (4, 'iv'),
             (1, 'i')],
)
register_style(
    'upper-roman',
    type='additive',
    range=(1, 4999),
    symbols=[(1000, 'M'), (900, 'CM'), (500, 'D'), (400, 'CD'),
             (100, 'C'), (90, 'XC'), (50, 'L'), (40, 'XL'),
             (10, 'X'), (9, 'IX'), (5, 'V'), (4, 'IV'),
             (1, 'I')],
)
register_style(
    'georgian',
    type='additive',
    range=(1, 19999),
    symbols=[
        (10000, 'ჵ'), (9000, 'ჰ'), (8000, 'ჯ'), (7000, 'ჴ'), (6000, 'ხ'),
        (5000, 'ჭ'), (4000, 'წ'), (3000, 'ძ'), (2000, 'ც'), (1000, 'ჩ'),
        (900, 'შ'), (800, 'ყ'), (700, 'ღ'), (600, 'ქ'),
        (500, 'ფ'), (400, 'ჳ'), (300, 'ტ'), (200, 'ს'), (100, 'რ'),
        (90, 'ჟ'), (80, 'პ'), (70, 'ო'), (60, 'ჲ'),
        (50, 'ნ'), (40, 'მ'), (30, 'ლ'), (20, 'კ'), (10, 'ი'),
        (9, 'თ'), (8, 'ჱ'), (7, 'ზ'), (6, 'ვ'),
        (5, 'ე'), (4, 'დ'), (3, 'გ'), (2, 'ბ'), (1, 'ა')],
)
register_style(
    'armenian',
    type='additive',
    range=(1, 9999),
    symbols=[
        (9000, 'Ք'), (8000, 'Փ'), (7000, 'Ւ'), (6000, 'Ց'),
        (5000, 'Ր'), (4000, 'Տ'), (3000, 'Վ'), (2000, 'Ս'), (1000, 'Ռ'),
        (900, 'Ջ'), (800, 'Պ'), (700, 'Չ'), (600, 'Ո'),
        (500, 'Շ'),(400, 'Ն'), (300, 'Յ'), (200, 'Մ'), (100, 'Ճ'),
        (90, 'Ղ'), (80, 'Ձ'), (70, 'Հ'), (60, 'Կ'),
        (50, 'Ծ'), (40, 'Խ'), (30, 'Լ'), (20, 'Ի'), (10, 'Ժ'),
        (9, 'Թ'), (8, 'Ը'), (7, 'Է'), (6, 'Զ'),
        (5, 'Ե'), (4, 'Դ'), (3, 'Գ'), (2, 'Բ'), (1, 'Ա')],
)
register_style(
    'lower-alpha',
    type='alphabetic',
    symbols='a b c d e f g h i j k l m n o p q r s t u v w x y z'.split(),
)
register_style(
    'upper-alpha',
    type='alphabetic',
    symbols='A B C D E F G H I J K L M N O P Q R S T U V W X Y Z'.split(),
)
register_style(
    'lower-greek',
    type='alphabetic',
    symbols='α β γ δ ε ζ η θ ι κ λ μ ν ξ ο π ρ σ τ υ φ χ ψ ω'.split()
)
register_style(
    'disc',
    type='repeating',
    symbols=['•'],  # U+2022, BULLET
    suffix='',
)
register_style(
    'circle',
    type='repeating',
    symbols=['◦'],  # U+25E6 WHITE BULLET
    suffix='',
)
register_style(
    'square',
    type='repeating',
    # CSS Lists 3 suggests U+25FE BLACK MEDIUM SMALL SQUARE
    # But I think this one looks better.
    symbols=['▪'],  # U+25AA BLACK SMALL SQUARE
    suffix='',
)
register_style(
    'lower-latin',
    type='override',
    override='lower-alpha',
)
register_style(
    'upper-latin',
    type='override',
    override='upper-alpha',
)


def format(value, counter_style):
    """
    Return a representation of ``value`` formatted by ``counter_style``
    or one of its fallback.

    The representation includes negative signs, but not the prefix and suffix.

    """
    if counter_style == 'none':
        return ''
    failed_styles = set()  # avoid fallback loops
    while True:
        if counter_style == 'decimal' or counter_style in failed_styles:
            return str(value)
        style = STYLES[counter_style]
        low, high = style['range']
        if low <= value <= high:
            representation = style['formatter'](value)
            if representation is not None:
                return representation
        failed_styles.add(counter_style)
        counter_style = style['fallback']


def format_list_marker(value, counter_style):
    """
    Return a representation of ``value`` formatted for a list marker.

    This is the same as :func:`format()`, but includes the counter’s
    prefix and suffix.
    """
    style = STYLES[counter_style]
    return style['prefix'] + format(value, counter_style) + style['suffix']
